
__all__ = ["MLRegressionExperiments"]

from ai4water.utils.utils import get_version_info, dateandtime_now
from ._main import Experiments
from .utils import regression_space


try:
    import catboost
except ModuleNotFoundError:
    catboost = None

try:
    import lightgbm
except ModuleNotFoundError:
    lightgbm = None

try:
    import xgboost
except ModuleNotFoundError:
    xgboost = None

import sklearn

VERSION_INFO = get_version_info(sklearn=sklearn)


class MLRegressionExperiments(Experiments):
    """
    Compares peformance of 40+ machine learning models for a regression problem.
    The experiment consists of `models` which are run using `fit()` method. A `model`
    is one experiment. 

    The user can define new `models` by subclassing this class. In fact any new
    method in the sub-class which starts with `model_` wll be considered
    as a new `model`. Otherwise the user has to overwite the attribute `models` to
    redefine, which methods (of class) are to be used as models and which should not. The
    method which is a `model` must only return key word arguments which will be
    streamed to the `Model` using `build_and_run` method. Inside this new method
    the user must define, which parameters to optimize, their param_space for optimization
    and the initial values to use for optimization.

    """

    def __init__(self,
                 param_space=None,
                 x0=None,
                 cases=None,
                 exp_name='MLRegressionExperiments',
                 num_samples=5,
                 verbosity=1,
                 **model_kwargs):
        """
        Initializes the class

        Arguments:
            param_space: dimensions of parameters which are to be optimized. These
                can be overwritten in `models`.
            x0 list: initial values of the parameters which are to be optimized.
                These can be overwritten in `models`
            exp_name str: name of experiment, all results will be saved within this folder
            model_kwargs dict: keyword arguments which are to be passed to `Model`
                and are not optimized.

        Examples:
            >>> from ai4water.datasets import busan_beach
            >>> from ai4water.experiments import MLRegressionExperiments
            >>> # first compare the performance of all available models without optimizing their parameters
            >>> data = busan_beach()  # read data file, in this case load the default data
            >>> inputs = list(data.columns)[0:-1]  # define input and output columns in data
            >>> outputs = list(data.columns)[-1]
            >>> comparisons = MLRegressionExperiments(
            ...       input_features=inputs, output_features=outputs,
            ...       nan_filler= {'method': 'KNNImputer', 'features': inputs} )
            >>> comparisons.fit(data=data,run_type="dry_run")
            >>> comparisons.compare_errors('r2')
            >>> # find out the models which resulted in r2> 0.5
            >>> best_models = comparisons.compare_errors('r2', cutoff_type='greater',
            ...                                                cutoff_val=0.3)
            >>> best_models = [m[1] for m in best_models.values()]
            >>> # now build a new experiment for best models and otpimize them
            >>> comparisons = MLRegressionExperiments(
            ...     inputs_features=inputs, output_features=outputs,
            ...     nan_filler= {'method': 'KNNImputer', 'features': inputs},
            ...     exp_name="BestMLModels")
            >>> comparisons.fit(data=data, run_type="optimize", include=best_models)
            >>> comparisons.compare_errors('r2')
            >>> comparisons.taylor_plot()  # see help(comparisons.taylor_plot()) to tweak the taylor plot

        """
        self.param_space = param_space
        self.x0 = x0
        self.model_kws = model_kwargs

        if exp_name == "MLRegressionExperiments":
            exp_name = f"{exp_name}_{dateandtime_now()}"

        super().__init__(cases=cases, exp_name=exp_name, num_samples=num_samples, verbosity=verbosity)

        self.regression_space = regression_space(num_samples=num_samples)

        if catboost is None:
            self.models.remove('model_CatBoostRegressor')
        if lightgbm is None:
            self.models.remove('model_LGBMRegressor')
        if xgboost is None:
            self.models.remove('model_XGBRFRegressor')

        sk_maj_ver = int(sklearn.__version__.split('.')[0])
        sk_min_ver = int(sklearn.__version__.split('.')[1])
        if sk_maj_ver == 0 and sk_min_ver < 23:
            for m in ['model_PoissonRegressor', 'model_TweedieRegressor']:
                self.models.remove(m)

    @property
    def tpot_estimator(self):
        try:
            from tpot import TPOTRegressor
        except (ModuleNotFoundError, ImportError):
            TPOTRegressor = None
        return TPOTRegressor

    @property
    def mode(self):
        return "regression"

    def model_AdaBoostRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostRegressor.html

        self.path = "sklearn.ensemble.AdaBoostRegressor"
        self.param_space = self.regression_space["AdaBoostRegressor"]["param_space"]
        self.x0 = self.regression_space["AdaBoostRegressor"]["x0"]

        return {'model': {'AdaBoostRegressor': kwargs}}

    def model_ARDRegression(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html

        self.path = "sklearn.linear_model.ARDRegression"
        self.param_space = self.regression_space["ARDRegression"]["param_space"]
        self.x0 = self.regression_space["ARDRegression"]["x0"]

        return {'model': {'ARDRegression': kwargs}}

    def model_BaggingRegressor(self, **kwargs):

        self.path = "sklearn.ensemble.BaggingRegressor"
        self.param_space = self.regression_space["BaggingRegressor"]["param_space"]
        self.x0 = self.regression_space["BaggingRegressor"]["x0"]

        return {'model': {'BaggingRegressor': kwargs}}

    def model_BayesianRidge(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html

        self.path = "sklearn.linear_model.BayesianRidge"
        self.param_space = self.regression_space["BayesianRidge"]["param_space"]
        self.x0 = self.regression_space["BayesianRidge"]["x0"]

        return {'model': {'BayesianRidge': kwargs}}

    def model_CatBoostRegressor(self, **kwargs):
        # https://catboost.ai/docs/concepts/python-reference_parameters-list.html

        self.path = "catboost.CatBoostRegressor"
        self.param_space = self.regression_space["CatBoostRegressor"]["param_space"]
        self.x0 = self.regression_space["CatBoostRegressor"]["x0"]

        return {'model': {'CatBoostRegressor': kwargs}}

    def model_DecisionTreeRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html

        self.path = "sklearn.tree.DecisionTreeRegressor"
        # TODO not converging
        self.param_space = self.regression_space["DecisionTreeRegressor"]["param_space"]
        self.x0 = self.regression_space["DecisionTreeRegressor"]["x0"]

        return {'model': {'DecisionTreeRegressor': kwargs}}

    def model_DummyRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.dummy.DummyRegressor.html

        self.path = "sklearn.dummy.DummyRegressor"
        self.param_space = self.regression_space["DummyRegressor"]["param_space"]
        self.x0 = self.regression_space["DummyRegressor"]["x0"]
        kwargs.update({'constant': 0.2,
                       'quantile': 0.2})

        return {'model': {'DummyRegressor': kwargs}}

    def model_ElasticNet(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html

        self.path = "sklearn.linear_model.ElasticNet"
        self.param_space = self.regression_space["ElasticNet"]["param_space"]
        self.x0 = self.regression_space["ElasticNet"]["x0"]

        return {'model': {'ElasticNet': kwargs}}

    def model_ElasticNetCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNetCV.html

        self.path = "sklearn.linear_model.ElasticNetCV"
        self.param_space = self.regression_space["ElasticNetCV"]["param_space"]
        self.x0 = self.regression_space["ElasticNetCV"]["x0"]

        return {'model': {'ElasticNetCV': kwargs}}

    def model_ExtraTreeRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.tree.ExtraTreeRegressor.htm

        self.path = "sklearn.tree.ExtraTreeRegressor"
        self.param_space = self.regression_space["ExtraTreeRegressor"]["param_space"]
        self.x0 = self.regression_space["ExtraTreeRegressor"]["x0"]

        return {'model': {'ExtraTreeRegressor': kwargs}}

    def model_ExtraTreesRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesRegressor.html

        self.path = "sklearn.ensemble.ExtraTreesRegressor"
        self.param_space = self.regression_space["ExtraTreesRegressor"]["param_space"]
        self.x0 = self.regression_space["ExtraTreesRegressor"]["x0"]

        return {'model': {'ExtraTreesRegressor': kwargs}}

    # def model_GammaRegressor(self, **kwargs):
    # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.GammaRegressor.html?highlight=gammaregressor
    #     self.param_space = [
    #         Real(low=0.0, high=1.0, name='alpha', num_samples=self.num_samples),
    #         Integer(low=50, high=500, name='max_iter', num_samples=self.num_samples),
    #         Real(low= 1e-6, high= 1e-2, name='tol', num_samples=self.num_samples),
    #         Categorical(categories=[True, False], name='warm_start'),
    #         Categorical(categories=[True, False], name='fit_intercept')
    #     ]
    #     self.x0 = [0.5, 100,1e-6, True, True]
    #     return {'model': {'GammaRegressor': kwargs}}

    def model_GaussianProcessRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessRegressor.html

        self.path = "sklearn.gaussian_process.GaussianProcessRegressor"
        self.param_space = self.regression_space["GaussianProcessRegressor"]["param_space"]
        self.x0 = self.regression_space["GaussianProcessRegressor"]["x0"]

        return {'model': {'GaussianProcessRegressor': kwargs}}

    def model_GradientBoostingRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingRegressor.html

        self.path = "sklearn.ensemble.GradientBoostingRegressor"
        self.param_space = self.regression_space["GradientBoostingRegressor"]["param_space"]
        self.x0 = self.regression_space["GradientBoostingRegressor"]["x0"]

        return {'model': {'GradientBoostingRegressor': kwargs}}

    def model_HistGradientBoostingRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.HistGradientBoostingRegressor.html

        # TODO not hpo not converging
        self.path = "sklearn.ensemble.HistGradientBoostingRegressor"
        self.param_space = self.regression_space["HistGradientBoostingRegressor"]["param_space"]
        self.x0 = self.regression_space["HistGradientBoostingRegressor"]["x0"]

        return {'model': {'HistGradientBoostingRegressor':kwargs}}

    def model_HuberRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.HuberRegressor.html

        self.path = "sklearn.linear_model.HuberRegressor"
        self.param_space = self.regression_space["HuberRegressor"]["param_space"]
        self.x0 = self.regression_space["HuberRegressor"]["x0"]

        return {'model': {'HuberRegressor': kwargs}}

    def model_KernelRidge(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html

        self.path = "sklearn.kernel_ridge.KernelRidge"
        self.param_space = self.regression_space["KernelRidge"]["param_space"]
        self.x0 = self.regression_space["KernelRidge"]["x0"]

        return {'model': {'KernelRidge': kwargs}}

    def model_KNeighborsRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsRegressor.html

        self.path = "sklearn.neighbors.KNeighborsRegressor"
        self.param_space = self.regression_space["KNeighborsRegressor"]["param_space"]
        self.x0 = self.regression_space["KNeighborsRegressor"]["x0"]

        return {'model': {'KNeighborsRegressor': kwargs}}

    def model_LassoLars(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html

        self.path = "sklearn.linear_model.LassoLars"
        self.param_space = self.regression_space["LassoLars"]["param_space"]
        self.x0 = self.regression_space["LassoLars"]["x0"]

        return {'model': {'LassoLars': kwargs}}

    def model_Lars(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lars.html

        self.path = "sklearn.linear_model.Lars"
        self.param_space = self.regression_space["Lars"]["param_space"]
        self.x0 = self.regression_space["Lars"]["x0"]

        return {'model': {'Lars': kwargs}}

    def model_LarsCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LarsCV.html

        self.path = "sklearn.linear_model.LarsCV"
        self.param_space = self.regression_space["LarsCV"]["param_space"]
        self.x0 = self.regression_space["LarsCV"]["x0"]

        return {'model': {'LarsCV': kwargs}}

    def model_LinearSVR(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVR.html

        self.path = "sklearn.svm.LinearSVR"
        self.param_space = self.regression_space["LinearSVR"]["param_space"]
        self.x0 = self.regression_space["LinearSVR"]["x0"]

        return {'model': {'LinearSVR': kwargs}}

    def model_Lasso(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html

        self.path = "sklearn.linear_model.Lasso"
        self.param_space = self.regression_space["Lasso"]["param_space"]
        self.x0 = self.regression_space["Lasso"]["x0"]

        return {'model': {'Lasso': kwargs}}

    def model_LassoCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html

        self.path = "sklearn.linear_model.LassoCV"
        self.param_space = self.regression_space["LassoCV"]["param_space"]
        self.x0 = self.regression_space["LassoCV"]["x0"]

        return {'model': {'LassoCV': kwargs}}

    def model_LassoLarsCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLarsCV.html

        self.path = "sklearn.linear_model.LassoLarsCV"
        self.param_space = self.regression_space["LassoLarsCV"]["param_space"]
        self.x0 = self.regression_space["LassoLarsCV"]["x0"]

        return {'model': {'LassoLarsCV': kwargs}}

    def model_LassoLarsIC(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLarsIC.html

        self.path = "sklearn.linear_model.LassoLarsIC"
        self.param_space = self.regression_space["LassoLarsIC"]["param_space"]
        self.x0 = self.regression_space["LassoLarsIC"]["x0"]

        return {'model': {'LassoLarsIC': kwargs}}

    def model_LGBMRegressor(self, **kwargs):
        # https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html

        self.path = "lightgbm.LGBMRegressor"
        self.param_space = self.regression_space["LGBMRegressor"]["param_space"]
        self.x0 = self.regression_space["LGBMRegressor"]["x0"]

        return {'model': {'LGBMRegressor': kwargs}}

    def model_LinearRegression(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html

        self.path = "sklearn.linear_model.LinearRegression"
        self.param_space = self.regression_space["LinearRegression"]["param_space"]
        self.x0 = self.regression_space["LinearRegression"]["x0"]

        return {'model': {'LinearRegression': kwargs}}

    def model_MLPRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html

        self.path = "sklearn.neural_network.MLPRegressor"
        self.param_space = self.regression_space["MLPRegressor"]["param_space"]
        self.x0 = self.regression_space["MLPRegressor"]["x0"]

        return {'model': {'MLPRegressor': kwargs}}

    def model_NuSVR(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html

        self.path = "sklearn.svm.NuSVR"
        self.param_space = self.regression_space["NuSVR"]["param_space"]
        self.x0 = self.regression_space["NuSVR"]["x0"]

        return {'model': {'NuSVR': kwargs}}

    def model_OrthogonalMatchingPursuit(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.OrthogonalMatchingPursuit.html

        self.path = "sklearn.linear_model.OrthogonalMatchingPursuit"
        self.param_space = self.regression_space["OrthogonalMatchingPursuit"]["param_space"]
        self.x0 = self.regression_space["OrthogonalMatchingPursuit"]["x0"]

        return {'model': {'OrthogonalMatchingPursuit': kwargs}}

    def model_OrthogonalMatchingPursuitCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.OrthogonalMatchingPursuitCV.html

        self.path = "sklearn.linear_model.OrthogonalMatchingPursuitCV"
        self.param_space = self.regression_space["OrthogonalMatchingPursuitCV"]["param_space"]
        self.x0 = self.regression_space["OrthogonalMatchingPursuitCV"]["x0"]

        return {'model': {'OrthogonalMatchingPursuitCV': kwargs}}

    def model_OneClassSVM(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.svm.OneClassSVM.html

        self.path = "sklearn.svm.OneClassSVM"
        self.param_space = self.regression_space["OneClassSVM"]["param_space"]
        self.x0 = self.regression_space["OneClassSVM"]["x0"]

        return {'model': {'OneClassSVM': kwargs}}

    def model_PoissonRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.PoissonRegressor.html

        self.path = "sklearn.linear_model.PoissonRegressor"
        self.param_space = self.regression_space["PoissonRegressor"]["param_space"]
        self.x0 = self.regression_space["PoissonRegressor"]["x0"]

        return {'model': {'PoissonRegressor': kwargs}}

    def model_Ridge(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html

        self.path = "sklearn.linear_model.Ridge"
        self.param_space = self.regression_space["Ridge"]["param_space"]
        self.x0 = self.regression_space["Ridge"]["x0"]

        return {'model': {'Ridge': kwargs}}

    def model_RidgeCV(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RidgeCV.html

        self.path = "sklearn.linear_model.RidgeCV"
        self.param_space = self.regression_space["RidgeCV"]["param_space"]
        self.x0 = self.regression_space["RidgeCV"]["x0"]

        return {'model': {'RidgeCV': kwargs}}

    def model_RadiusNeighborsRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.RadiusNeighborsRegressor.html

        self.path = "sklearn.neighbors.RadiusNeighborsRegressor"
        self.param_space = self.regression_space["RadiusNeighborsRegressor"]["param_space"]
        self.x0 = self.regression_space["RadiusNeighborsRegressor"]["x0"]

        return {'model': {'RadiusNeighborsRegressor': kwargs}}

    def model_RANSACRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RANSACRegressor.html

        self.path = "sklearn.linear_model.RANSACRegressor"
        self.param_space = self.regression_space["RANSACRegressor"]["param_space"]
        self.x0 = self.regression_space["RANSACRegressor"]["x0"]

        return {'model': {'RANSACRegressor': kwargs}}

    def model_RandomForestRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html

        self.path = "sklearn.ensemble.RandomForestRegressor"
        self.param_space = self.regression_space["RandomForestRegressor"]["param_space"]
        self.x0 = self.regression_space["RandomForestRegressor"]["x0"]

        return {'model': {'RandomForestRegressor': kwargs}}

    def model_SVR(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html

        self.path = "sklearn.svm.SVR"
        self.param_space = self.regression_space["SVR"]["param_space"]
        self.x0 = self.regression_space["SVR"]["x0"]

        return {'model': {'SVR': kwargs}}

    def model_SGDRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html

        self.path = "sklearn.linear_model.SGDRegressor"
        self.param_space = self.regression_space["SGDRegressor"]["param_space"]
        self.x0 = self.regression_space["SGDRegressor"]["x0"]

        return {'model': {'SGDRegressor': kwargs}}

    # def model_TransformedTargetRegressor(self, **kwargs):
    #     ## https://scikit-learn.org/stable/modules/generated/sklearn.compose.TransformedTargetRegressor.html
    #     self.param_space = [
    #         Categorical(categories=[None], name='regressor'),
    #         Categorical(categories=[None], name='transformer'),
    #         Categorical(categories=[None], name='func')
    #     ]
    #     self.x0 = [None, None, None]
    #     return {'model': {'TransformedTargetRegressor': kwargs}}

    def model_TweedieRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TweedieRegressor.html

        self.path = "sklearn.linear_model.TweedieRegressor"
        self.param_space = self.regression_space["TweedieRegressor"]["param_space"]
        self.x0 = self.regression_space["TweedieRegressor"]["x0"]

        return {'model': {'TweedieRegressor': kwargs}}

    def model_TheilsenRegressor(self, **kwargs):
        # https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TheilSenRegressor.html

        self.path = "sklearn.linear_model.TheilSenRegressor"
        self.param_space = self.regression_space["TheilSenRegressor"]["param_space"]
        self.x0 = self.regression_space["TheilSenRegressor"]["x0"]

        return {'model': {'TheilSenRegressor': kwargs}}

    # TODO
    # def model_GAMMAREGRESSOR(self, **kwargs):
    #     # ValueError: Some value(s) of y are out of the valid range for family GammaDistribution
    #     return {'GAMMAREGRESSOR': {}}

    def model_XGBRFRegressor(self, **kwargs):
        # https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRFRegressor

        self.path = "xgboost.XGBRFRegressor"
        self.param_space = self.regression_space["XGBRFRegressor"]["param_space"]
        self.x0 = self.regression_space["XGBRFRegressor"]["x0"]

        return {'model': {'XGBRFRegressor': kwargs}}

    def model_XGBRegressor(self, **kwargs):
        # ##https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRegressor

        self.path = "xgboost.XGBRegressor"
        self.param_space = self.regression_space["XGBRegressor"]["param_space"]
        self.x0 = self.regression_space["XGBRegressor"]["x0"]

        return {'model': {'XGBRegressor': kwargs}}



"""
main models are pytorch based and Model.
All tensorflow based models can be implemented purely using Model.
"""
from ai4water.main import Model

try:
    from ai4water.pytorch_models import IMVModel
    from ai4water.pytorch_models import HARHNModel
except AttributeError:
    print("\n{}Pytorch models could not be imported {}\n".format(10*'*', 10*'*'))

try:
    from ai4water.tf_models import InputAttentionModel
    from ai4water.tf_models import DualAttentionModel
except AttributeError:
    print("\n{}Tensorflow models could not be imported {}\n".format(10 * '*', 10 * '*'))

__version__ = "1.0b4"



import os
import random
import warnings
from collections import OrderedDict

try:
    import imageio
except ModuleNotFoundError:
    imageio = None


try:
    import shapefile
except ModuleNotFoundError:
    shapefile = None

import matplotlib.pyplot as plt
import numpy as np


MSG = f"""
        If you installed shapely using pip, try to resintall it 
        (after uninstalling the previous installtin obviously)
        by manually downloading the wheel file from 
        https://www.lfd.uci.edu/~gohlke/pythonlibs/#shapely
        and then istalling using the wheel file using following command
        pip install path/to/wheel.whl"""

M2ToAcre = 0.0002471     # meter square to Acre

COLORS = ['#CDC0B0', '#00FFFF', '#76EEC6', '#C1CDCD', '#E3CF57', '#EED5B7', '#8B7D6B', '#0000FF', '#8A2BE2', '#9C661F',
          '#FF4040', '#8A360F', '#98F5FF', '#FF9912', '#B23AEE', '#9BCD9B', '#8B8B00']


def get_sorted_dict(dictionary):
    """sorts the dictionary based on its keys and returns the new dictionary named sorted_dict"""
    sorted_dict = OrderedDict()
    for k in sorted(dictionary):
        sorted_dict[k] = dictionary[k]
    return sorted_dict


def get_areas_geoms(shp_reader):
    """returns lists containing areas of all records in shapefile and geometries of all records
     in shape file and number of records in shapefile"""

    try:
        from shapely.geometry import shape
    except FileNotFoundError:
        raise FileNotFoundError(MSG)

    except OSError:
        warnings.warn(MSG, UserWarning)
        shape = None  # so that docs can be built
    except ModuleNotFoundError:
        warnings.warn(MSG, UserWarning)
        shape = None  # so that docs can be built

    shapes = shp_reader.shapes()

    geometries = [None] * len(shapes)  # a container for geometries of shapefile
    areas = [None] * len(shapes)  # a container for areas of shapefile
    for shp in range(len(shapes)):
        feature = shp_reader.shapeRecords()[shp]  # pyshp
        first = feature.shape.__geo_interface__  # pyshp
        geometries[shp] = shape(first)  # pyshp to shapely  geometry
        areas[shp] = shape(first).area * M2ToAcre

    return areas, geometries


def check_shp_validity(geom_list, no_of_lus, name='landuse', verbosity=1):
    new_geom_list = [None] * no_of_lus  # a container for geometries of landuse shapefile with corrected topology
    for lu in range(no_of_lus):  # iterating over each landuse
        luf = geom_list[lu]
        if not luf.is_valid:
            n = 0
            for j in range(len(luf)):  # iterating over each ring of feature which is invalid
                if not luf[j].is_valid:  # checking which ring in feature is invalid
                    n = n + 1  # counting number of invalid rings in a feature
            new_geom_list[lu] = luf.buffer(0)  # correcting the 'self-intersection' for the feature which is invalid
        else:
            new_geom_list[lu] = luf

    # checking the validity of each landuse once again to make sure that each landuse's is valid
    for lu in range(no_of_lus):
        sub_lu = new_geom_list[lu]
        if sub_lu.is_valid and verbosity:
            print('{} {} is valid now'.format(name, lu))
        elif verbosity:
            print('{} {} is still invalid'.format(name, lu))
    return new_geom_list


def get_total_area(file_to_read):
    shape_area = 0.0
    for sub_shp in file_to_read:
        shape_area += sub_shp.area*M2ToAcre
    return shape_area


class GifUtil(object):

    def __init__(self, folder, initials=None, contains=None):
        self.init = initials  # starting name of files
        self.contains = contains
        self.input_files = []   # container for input files
        self.get_all_files(initials)   # get files to make gif
        self.dir = folder  # folder containing input files and output gif

    def get_all_files(self, init):
        for file in os.listdir("./plots"):
            if file.endswith(".png"):
                if self.init:
                    if file.startswith(init):
                        self.input_files.append(file)
                if self.contains:
                    if self.contains in file:
                        self.input_files.append(file)

    def make_gif(self, duration=0.5, name=None):
        if name is None:
            if self.init:
                name = self.init
            else:
                name = self.contains

        images = []
        for file in np.sort(self.input_files):
            filename = os.path.join(self.dir, file)
            images.append(imageio.imread(filename))
        imageio.mimsave(os.path.join(self.dir, name + '.gif'), images, duration=duration)

    def remove_images(self):
        for img in self.input_files:
            path = os.path.join(self.dir, img)
            if os.path.exists(path):
                os.remove(path)


def find_records(shp_file, record_name, feature_number):
    """find the metadata about feature given its feature number and column_name which contains the data"""
    assert os.path.exists(shp_file), f'{shp_file} does not exist'
    shp_reader = shapefile.Reader(shp_file)
    col_no = find_col_name(shp_reader, record_name)

    if col_no == -99:
        raise ValueError(f'no column named {record_name} found in {shp_reader.shapeName}')
    else:
        # print(col_no, 'is the col no')
        name = get_record_in_col(shp_reader, feature_number, col_no)
    return name


def find_col_name(shp_reader, field_name):
    _col_no = 0
    col_no = -99
    for fields in shp_reader.fields:
        _col_no += 1
        for field in fields:
            if field == field_name:
                col_no = _col_no
                break
    return col_no


def get_record_in_col(shp_reader, i, col_no):
    recs = shp_reader.records()
    col_no = col_no - 2  # -2, 1 for index reduction, 1 for a junk column shows up in records
    return recs[i][col_no]


def plot_shapefile(shp_files,
                   labels=None,
                   show_all_together=True,
                   bbox_shp=None,
                   recs=None, rec_idx=None,
                   leg_kws=None,
                   save=False,
                   colors=None,
                   markersize=12,
                   save_kws=None):

    """
    leg_kws:{'bbox_to_anchor': (1.02, -0.15),
                   'numpoints': 1,
                   'fontsize': 16,
                   'markerscale':2}
    save_kws:{'fname': 'point_plot', 'bbox_inches': 'tight'}
    """
    if not isinstance(shp_files, list):
        shp_files = [shp_files]

    if leg_kws is None:
        leg_kws = {'bbox_to_anchor': (0.93, -0.15),
                   'numpoints': 1,
                   'fontsize': 16,
                   'markerscale': 2}
    if labels is None:
        labels = {}
    if save_kws is None:
        save_kws = {'fname': 'point_plot', 'dpi': 300, 'bbox_inches': 'tight'}

    records = shapefile.Reader(shp_files[0]).shapeRecords()
    Colors = random.choices(COLORS, k=len(records))

    if len(shp_files) > 1:
        for i in range(1, len(shp_files)):
            shp_reader = shapefile.Reader(shp_files[i])
            records += shp_reader.shapeRecords()
            Colors += random.choices(COLORS, k=len(shp_reader.shapeRecords()))

    plt.close('all')
    for feature, n in zip(records, Colors):

        if recs is not None:
            assert isinstance(rec_idx, int)
            rec = feature.record[rec_idx]
        else:
            rec, recs = '', ''

        if rec in recs:
            f_if = feature.shape.__geo_interface__
            if f_if is None:
                pass
            else:
                if f_if['type'].lower() in ['point']:  # it is point
                    c = colors.get(rec, random.choice(COLORS))
                    plt.plot(*f_if['coordinates'], '*', label=labels.get(rec, rec), color=c, markersize=markersize)
                else:
                    plot_polygon_feature(feature, n, shapefile.Reader(shp_files[0]).bbox)

    if bbox_shp is not None:
        shp_reader = shapefile.Reader(bbox_shp)
        records = shp_reader.shapeRecords()
        for feature, n in zip(records, Colors):
            plot_polygon_feature(feature, n, shapefile.Reader(shp_files[0]).bbox)

        plt.legend(**leg_kws)
        if not show_all_together:
            plt.show()

    if save:
        plt.savefig(**save_kws)
    # if show_all_together:
    plt.show()
    # shp_reader.close()
    return


def plot_polygon_feature(feature, n, bbox):
    f_if = feature.shape.__geo_interface__
    polys = len(f_if['coordinates'])
    def_col = n
    for i in range(polys):
        a = np.array(f_if['coordinates'][i])
        if a.ndim < 2 and len(a.shape) > 0:
            c = a
            m = max([len(ci) for ci in c])
            for ci in c:
                col = 'k' if len(ci) != m else def_col
                x = np.array([k[0] for k in ci])
                y = np.array([k[1] for k in ci])
                plt.plot(x, y, col, label="__none__", linewidth=0.5)

        elif len(a.shape) > 0:
            b = a.reshape(-1, 2)
            plt.plot(b[:, 0], b[:, 1], def_col)
        plt.ylim([bbox[1], bbox[3]])
        plt.xlim([bbox[0], bbox[2]])
    return




from ai4water.backend import sklearn_models


def convert_ai4water_model(old_model, framework=None, explainer=None):
    """convert ai4water's Model class to sklearn/xgboost..etc type model classes
    """
    new_model = old_model

    model_name = old_model.__class__.__name__

    if old_model.__class__.__name__ == "Model" and "ai4water" in str(type(old_model)):
        # this is ai4water model class
        if old_model.category == "ML":
            model_name = list(old_model.config['model'].keys())[0]
            new_model, _explainer = to_native(old_model, model_name)
            explainer = explainer or _explainer
            framework = "ML"
        else:
            framework = "DL"
            explainer = explainer or "DeepExplainer"
            if 'functional' in str(type(old_model)):
                new_model = functional_to_keras(old_model)

    return new_model, framework, explainer, model_name


def to_native(model, model_name:str):
    explainer = "TreeExplainer"

    if model_name.startswith("XGB"):
        import xgboost
        BaseModel = xgboost.__dict__[model_name]
    elif model_name.startswith("LGB"):
        import lightgbm
        BaseModel = lightgbm.__dict__[model_name]
    elif model_name.startswith("Cat"):
        import catboost
        BaseModel = catboost.__dict__[model_name]
    elif model_name in sklearn_models:
        BaseModel = sklearn_models[model_name]
        explainer = "KernelExplainer"

    else:
        raise ValueError

    class DummyModel(BaseModel):
        """First priority is to get attribute from ai4water's Model and then from
        the underlying library's model class."""
        def __getattribute__(self, item):
            return getattr(model, item)

        def __getattr__(self, item):
            return getattr(model._model, item)

    return DummyModel(), explainer


def get_features(features, features_to_explain):

    if features_to_explain is not None:
        if isinstance(features_to_explain, str):
            features_to_explain = [features_to_explain]
    else:
        features_to_explain = features

    assert isinstance(features_to_explain, list)

    for f in features_to_explain:
        assert f in features

    return features_to_explain


def functional_to_keras(old_model):
    """converts the model of functional api to keras model"""
    assert old_model.config['x_transformation'] is None
    assert old_model.config['y_transformation'] is None

    from tensorflow.keras.models import Model
    from tensorflow.keras.layers import Flatten

    # keras model from functional api
    old_model = old_model._model

    old_m_outputs = old_model.outputs
    if isinstance(old_m_outputs, list):
        assert len(old_m_outputs) == 1
        old_m_outputs = old_m_outputs[0]

    if len(old_m_outputs.shape) > 2:  # (None, ?, ?)
        new_outputs = Flatten()(old_m_outputs)  # (None, ?)
        assert new_outputs.shape.as_list()[-1] == 1  # (None, 1)
        new_model = Model(old_model.inputs, new_outputs)

    else:  # (None, ?)
        assert old_m_outputs.shape.as_list()[-1] == 1  # (None, 1)
        new_model = old_model

    return new_model



from typing import Any, Dict

import numpy as np
import pandas as pd
from scipy import linalg
import matplotlib.pyplot as plt


def auto_corr(x, nlags, demean=True):
    """
    autocorrelation like statsmodels
    https://stackoverflow.com/a/51168178
    """

    var = np.var(x)

    if demean:
        x -= np.mean(x)

    corr = np.full(nlags+1, np.nan, np.float64)
    corr[0] = 1.

    for lag in range(1, nlags+1):
        corr[lag] = np.sum(x[lag:]*x[:-lag])/len(x)/var

    return corr


def pac_yw(x, nlags):
    """partial autocorrelation according to ywunbiased method"""

    pac = np.full(nlags+1, fill_value=np.nan, dtype=np.float64)
    pac[0] = 1.

    for lag in range(1, nlags+1):
        pac[lag] = ar_yw(x, lag)[-1]

    return pac


def ar_yw(x, order=1, adj_needed=True, demean=True):
    """Performs autoregressor using Yule-Walker method.
    Returns:
        rho : np array
        coefficients of AR
    """
    x = np.array(x, dtype=np.float64)

    if demean:
        x -= x.mean()

    n = len(x)
    r = np.zeros(order+1, np.float64)
    r[0] = (x ** 2).sum() / n
    for k in range(1, order+1):
        r[k] = (x[0:-k] * x[k:]).sum() / (n - k * adj_needed)
    R = linalg.toeplitz(r[:-1])

    rho = np.linalg.solve(R, r[1:])
    return rho


def plot_autocorr(
        x,
        axis=None,
        plot_marker=True,
        show=True,
        legend=None,
        title=None,
        xlabel=None,
        vlines_colors=None,
        hline_color=None,
        marker_color=None,
        legend_fs=None
):

    if not axis:
        _, axis = plt.subplots()

    if plot_marker:
        axis.plot(x, 'o', color=marker_color, label=legend)
        if legend:
            axis.legend(fontsize=legend_fs)
    axis.vlines(range(len(x)), [0], x, colors=vlines_colors)
    axis.axhline(color=hline_color)

    if title:
        axis.set_title(title)
    if xlabel:
        axis.set_xlabel("Lags")

    if show:
        plt.show()

    return axis


def ccovf_np(x, y, unbiased=True, demean=True):
    n = len(x)
    if demean:
        xo = x - x.mean()
        yo = y - y.mean()
    else:
        xo = x
        yo = y
    if unbiased:
        xi = np.ones(n)
        d = np.correlate(xi, xi, 'full')
    else:
        d = n
    return (np.correlate(xo, yo, 'full') / d)[n - 1:]


def ccf_np(x, y, unbiased=True):
    """cross correlation between two time series
    # https://stackoverflow.com/a/24617594
    """
    cvf = ccovf_np(x, y, unbiased=unbiased, demean=True)
    return cvf / (np.std(x) * np.std(y))


def _missing_vals(data: pd.DataFrame) -> Dict[str, Any]:
    """
    Modified after https://github.com/akanz1/klib/blob/main/klib/utils.py#L197
     Gives metrics of missing values in the dataset.
    Parameters
    ----------
    data : pd.DataFrame
        2D dataset that can be coerced into Pandas DataFrame
    Returns
    -------
    Dict[str, float]
        mv_total: float, number of missing values in the entire dataset
        mv_rows: float, number of missing values in each row
        mv_cols: float, number of missing values in each column
        mv_rows_ratio: float, ratio of missing values for each row
        mv_cols_ratio: float, ratio of missing values for each column
    """

    data = pd.DataFrame(data).copy()
    mv_rows = data.isna().sum(axis=1)
    mv_cols = data.isna().sum(axis=0)
    mv_total = data.isna().sum().sum()
    mv_rows_ratio = mv_rows / data.shape[1]
    mv_cols_ratio = mv_cols / data.shape[0]

    return {
        "mv_total": mv_total,
        "mv_rows": mv_rows,
        "mv_cols": mv_cols,
        "mv_rows_ratio": mv_rows_ratio,
        "mv_cols_ratio": mv_cols_ratio,
    }


# This file contains code mostly or 90% from optuna library. The reason to rewrite this code is because
# I was unable to create an Optuna Study instance without `Storage` attribute. But it appears that
# we can calculate parameter importance without Storage attribute from Study instance. Thus
# the importance is calculated from Study instance without it having Storage attribute.
# also support to handle nan values in target array is added.
# The optuna library comes with following MIT licence.
"""
MIT License

Copyright (c) 2018 Preferred Networks, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import warnings

from collections import OrderedDict
from typing import Callable
from typing import List
from typing import Optional

import numpy as np

from optuna.logging import get_logger
from optuna._transform import _SearchSpaceTransform
from optuna.study import Study
from optuna.trial import FrozenTrial
from optuna.trial import TrialState
from optuna.importance._fanova import FanovaImportanceEvaluator
from optuna.visualization._plotly_imports import _imports
from optuna.visualization._utils import _check_plot_args
from optuna.visualization._plotly_imports import go
from optuna.importance import get_param_importances


from easy_mpl import bar_chart


logger = get_logger(__name__)



def _get_distributions(study, params):
    # based on supposition that get_distributions only returns an ordered dictionary and requiring `storage` attribute
    # of study is redundant
    assert params is None
    trial = study.trials[0]
    return OrderedDict(trial.distributions)


class ImportanceEvaluator(FanovaImportanceEvaluator):

    def evaluate(
        self,
        study: Study,
        params: Optional[List[str]] = None,
        *,
        target: Optional[Callable[[FrozenTrial], float]] = None,
    ):

        if target is None and study._is_multi_objective():
            raise ValueError(
                "If the `study` is being used for multi-objective optimization, "
                "please specify the `target`."
            )

        distributions = _get_distributions(study, params)
        if len(distributions) == 0:
            return OrderedDict()

        trials = []
        for trial in study.trials:
            if trial.state != TrialState.COMPLETE:
                continue
            if any(name not in trial.params for name in distributions.keys()):
                continue
            trials.append(trial)

        trans = _SearchSpaceTransform(distributions, transform_log=False, transform_step=False)

        n_trials = len(trials)
        trans_params = np.empty((n_trials, trans.bounds.shape[0]), dtype=np.float64)
        trans_values = np.empty(n_trials, dtype=np.float64)

        for trial_idx, trial in enumerate(trials):
            trans_params[trial_idx] = trans.transform(trial.params)
            trans_values[trial_idx] = trial.value if target is None else target(trial)

        # if nan values are present in target, use mean to fill them
        nan_idx = np.isnan(trans_values)
        if nan_idx.any():
            warnings.warn("Invalid value encountered in target values",
                UserWarning)
            # fill nan values with mean
            trans_values[nan_idx] = np.nanmean(trans_values)

        trans_bounds = trans.bounds
        column_to_encoded_columns = trans.column_to_encoded_columns

        if trans_params.size == 0:  # `params` were given but as an empty list.
            return OrderedDict()

        # Many (deep) copies of the search spaces are required during the tree traversal and using
        # Optuna distributions will create a bottleneck.
        # Therefore, search spaces (parameter distributions) are represented by a single
        # `numpy.ndarray`, coupled with a list of flags that indicate whether they are categorical
        # or not.

        evaluator = self._evaluator
        evaluator.fit(
            X=trans_params,
            y=trans_values,
            search_spaces=trans_bounds,
            column_to_encoded_columns=column_to_encoded_columns,
        )

        importances = {}
        variance = {}
        for i, name in enumerate(distributions.keys()):
            _mean, _std = evaluator.get_importance((i,))
            importances[name] = _mean
            variance[name] = {'mean': _mean, 'std': _std}

        total_importance = sum(importances.values())
        for name in importances:
            importances[name] /= total_importance

        sorted_importances = OrderedDict(
            reversed(
                sorted(importances.items(), key=lambda name_and_importance: name_and_importance[1])
            )
        )

        return sorted_importances, variance


def plot_param_importances(
    study: Study,
    evaluator = None,
    params: Optional[List[str]] = None,
    *,
    target: Optional[Callable[[FrozenTrial], float]] = None,
    target_name: str = "Objective Value",
):

    _imports.check()
    _check_plot_args(study, target, target_name)

    layout = go.Layout(
        title="Hyperparameter Importances",
        xaxis={"title": f"Importance for {target_name}"},
        yaxis={"title": "Hyperparameter"},
        showlegend=False,
    )

    # Importances cannot be evaluated without completed trials.
    # Return an empty figure for consistency with other visualization functions.
    trials = [trial for trial in study.trials if trial.state == TrialState.COMPLETE]
    if len(trials) == 0:
        logger.warning("Study instance does not contain completed trials.")
        return go.Figure(data=[], layout=layout)

    if evaluator is None:
        evaluator = ImportanceEvaluator()
    try:
        importances, importance_paras = get_param_importances(
            study, evaluator=evaluator, params=params, target=target
        )
    except RuntimeError:  # sometimes it is returning error e.g. when number of trials are < 4
        return None, None, None

    importances = OrderedDict(reversed(list(importances.items())))
    importance_values = list(importances.values())
    param_names = list(importances.keys())

    ax = bar_chart(importance_values, param_names, orient='h', show=False,
                   title="fANOVA hyperparameter importance",
                   xlabel="Relative Importance")

    return importances, importance_paras, ax


def _get_distribution(param_name: str, study: Study):
    for trial in study.trials:
        if param_name in trial.distributions:
            return trial.distributions[param_name]
    assert False






import json

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from ai4water.utils.utils import print_something


def consider_intervals(data, intervals):
    _source = data
    if intervals is not None:
        if isinstance(data, pd.DataFrame):
            try:  # if indices in intervals are of same type as that of index
                # -1 so that .loc and .iloc give same results, however this is not possible
                # with DatetimeIndex
                if isinstance(data.index, pd.DatetimeIndex):
                    _source = pd.concat([data.loc[st:en] for st, en in intervals])
                else:
                    _source = pd.concat([data.loc[st:en - 1] for st, en in intervals])
            except TypeError:  # assuming indices in intervals are integers
                _source = pd.concat([data.iloc[st:en] for st, en in intervals])

    return _source


def load_data_from_hdf5(data_type, data):
    import h5py

    f = h5py.File(data, mode='r')

    g = f[data_type]
    weight_names = list(g.keys())
    weight_values = [np.asarray(g[weight_name]) for weight_name in weight_names]

    f.close()

    return weight_values


def check_for_classification(label: np.ndarray, to_categorical):

    assert isinstance(label, np.ndarray), f"""
                            classification problem for label of type {label.__class__.__name__} not implemented yet"""

    # for clsasification, it should be 2d
    label = label.reshape(-1, label.shape[1])
    if to_categorical:
        assert label.shape[1] == 1
        label = OneHotEncoder(sparse=False).fit_transform(label)
    # else:   # mutlti_label/binary problem
    #     # todo, is only binary_crossentropy is binary/multi_label problem?
    #     pass #assert self.loss_name() in ['binary_crossentropy']
    return label


def decode(json_string):
    return json.loads(json_string, object_hook=_decode_helper)


def _decode_helper(obj):
    """A decoding helper that is TF-object aware."""
    if isinstance(obj, dict) and 'class_name' in obj:

        if obj['class_name'] == '__tuple__':
            return tuple(_decode_helper(i) for i in obj['items'])
        elif obj['class_name'] == '__ellipsis__':
            return Ellipsis
    return obj


import tensorflow as tf
from tensorflow.keras import layers

from .attention_layers import ChannelAttention, SpatialAttention, regularized_padded_conv


def _get_tensor_shape(t):
    return t.shape


class ConditionalRNN(tf.keras.layers.Layer):

    # Arguments to the RNN like return_sequences, return_state...
    def __init__(self, units,
                 activation='tanh',
                 recurrent_activation='sigmoid',
                 use_bias=True,
                 dropout=0.0,
                 recurrent_dropout=0.0,
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 cell=tf.keras.layers.LSTMCell, *args,
                 **kwargs):
        """
        Conditional RNN. Conditions time series on categorical data.
        :param units: int, The number of units in the RNN Cell
        :param cell: string, cell class or object (pre-instantiated). In the case of string, 'GRU',
        'LSTM' and 'RNN' are supported.
        :param args: Any parameters of the tf.keras.layers.RNN class, such as return_sequences,
        return_state, stateful, unroll...
        """
        super().__init__()
        self.units = units
        self.final_states = None
        self.init_state = None
        if isinstance(cell, str):
            if cell.upper() == 'GRU':
                cell = tf.keras.layers.GRUCell
            elif cell.upper() == 'LSTM':
                cell = tf.keras.layers.LSTMCell
            elif cell.upper() == 'RNN':
                cell = tf.keras.layers.SimpleRNNCell
            else:
                raise Exception('Only GRU, LSTM and RNN are supported as cells.')
        self._cell = cell if hasattr(cell, 'units') else cell(units=units,
                                                              activation=activation,
                                                              dropout=dropout,
                                                              recurrent_dropout=recurrent_dropout,
                                                              recurrent_activation=recurrent_activation,
                                                              kernel_initializer=kernel_regularizer,
                                                              recurrent_regularizer=recurrent_regularizer,
                                                              use_bias=use_bias
                                                              )
        self.rnn = tf.keras.layers.RNN(cell=self._cell, *args, **kwargs)

        # single cond
        self.cond_to_init_state_dense_1 = tf.keras.layers.Dense(units=self.units)

        # multi cond
        max_num_conditions = 10
        self.multi_cond_to_init_state_dense = []

        for _ in range(max_num_conditions):
            self.multi_cond_to_init_state_dense.append(tf.keras.layers.Dense(units=self.units))

        self.multi_cond_p = tf.keras.layers.Dense(1, activation=None, use_bias=True)

    def _standardize_condition(self, initial_cond):
        initial_cond_shape = initial_cond.shape
        if len(initial_cond_shape) == 2:
            initial_cond = tf.expand_dims(initial_cond, axis=0)
        first_cond_dim = initial_cond.shape[0]
        if isinstance(self._cell, tf.keras.layers.LSTMCell):
            if first_cond_dim == 1:
                initial_cond = tf.tile(initial_cond, [2, 1, 1])
            elif first_cond_dim != 2:
                raise Exception('Initial cond should have shape: [2, batch_size, hidden_size] '
                                'or [batch_size, hidden_size]. Shapes do not match.', initial_cond_shape)
        elif isinstance(self._cell, tf.keras.layers.GRUCell) or isinstance(self._cell, tf.keras.layers.SimpleRNNCell):
            if first_cond_dim != 1:
                raise Exception('Initial cond should have shape: [1, batch_size, hidden_size] '
                                'or [batch_size, hidden_size]. Shapes do not match.', initial_cond_shape)
        else:
            raise Exception('Only GRU, LSTM and RNN are supported as cells.')
        return initial_cond

    def __call__(self, inputs, *args, **kwargs):
        """
        :param inputs: List of n elements:
                    - [0] 3-D Tensor with shape [batch_size, time_steps, input_dim]. The inputs.
                    - [1:] list of tensors with shape [batch_size, cond_dim]. The conditions.
        In the case of a list, the tensors can have a different cond_dim.
        :return: outputs, states or outputs (if return_state=False)
        """
        assert (isinstance(inputs, list) or isinstance(inputs, tuple)) and len(inputs) >= 2, f"{type(inputs)}"
        x = inputs[0]
        cond = inputs[1:]
        if len(cond) > 1:  # multiple conditions.
            init_state_list = []
            for ii, c in enumerate(cond):
                init_state_list.append(self.multi_cond_to_init_state_dense[ii](self._standardize_condition(c)))
            multi_cond_state = self.multi_cond_p(tf.stack(init_state_list, axis=-1))
            multi_cond_state = tf.squeeze(multi_cond_state, axis=-1)
            self.init_state = tf.unstack(multi_cond_state, axis=0)
        else:
            cond = self._standardize_condition(cond[0])
            if cond is not None:
                self.init_state = self.cond_to_init_state_dense_1(cond)
                self.init_state = tf.unstack(self.init_state, axis=0)
        out = self.rnn(x, initial_state=self.init_state, *args, **kwargs)
        if self.rnn.return_state:
            outputs, h, c = out
            final_states = tf.stack([h, c])
            return outputs, final_states
        else:
            return out


class BasicBlock(layers.Layer):
    """
    The official implementation is at https://github.com/Jongchan/attention-module/blob/master/MODELS/cbam.py
    The implementation of [1] does not have two conv and bn paris. They just applied channel attention followed by
    spatial attention on inputs.

    [1] https://github.com/kobiso/CBAM-tensorflow/blob/master/attention_module.py#L39
    """
    expansion = 1

    def __init__(self, conv_dim, out_channels=32, stride=1, **kwargs):
        super(BasicBlock, self).__init__(**kwargs)

        # 1. BasicBlock模块中的共有2个卷积;BasicBlock模块中的第1个卷积层;
        self.conv1 = regularized_padded_conv(conv_dim, out_channels, kernel_size=3, strides=stride)
        self.bn1 = layers.BatchNormalization()

        # 2. 第2个；第1个卷积如果做stride就会有一个下采样，在这个里面就不做下采样了。这一块始终保持size一致，把stride固定为1
        self.conv2 = regularized_padded_conv(conv_dim, out_channels, kernel_size=3, strides=1)
        self.bn2 = layers.BatchNormalization()
        # ############################## 注意力机制 ###############################
        self.ca = ChannelAttention(conv_dim=conv_dim, in_planes=out_channels)
        self.sa = SpatialAttention(conv_dim=conv_dim)

        # # 3. 判断stride是否等于1,如果为1就是没有降采样。
        # if stride != 1 or in_channels != self.expansion * out_channels:
        #     self.shortcut = Sequential([regularized_padded_conv(self.expansion * out_channels,
        #                                                         kernel_size=1, strides=stride),
        #                                 layers.BatchNormalization()])
        # else:
        #     self.shortcut = lambda x, _: x

    def call(self, inputs, training=False):
        out = self.conv1(inputs)
        out = self.bn1(out, training=training)
        out = tf.nn.relu(out)

        out = self.conv2(out)
        out = self.bn2(out, training=training)
        # ############################## 注意力机制 ###############################
        out = self.ca(out) * out
        out = self.sa(out) * out

        # out = out + self.shortcut(inputs, training)
        # out = tf.nn.relu(out)

        return out


class scaled_dot_product_attention(layers.Layer):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, q, k, v, mask):
        """Calculate the attention weights.
        q, k, v must have matching leading dimensions.
        k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.
        The mask has different shapes depending on its type(padding or look ahead)
        but it must be broadcastable for addition.

        Args:
        q: query shape == (..., seq_len_q, depth)
        k: key shape == (..., seq_len_k, depth)
        v: value shape == (..., seq_len_v, depth_v)
        mask: Float tensor with shape broadcastable
              to (..., seq_len_q, seq_len_k). Defaults to None.

        Returns:
        output, attention_weights
        """

        matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)

        # scale matmul_qk
        dk = tf.cast(tf.shape(k)[-1], tf.float32)
        scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

        # add the mask to the scaled tensor.
        if mask is not None:
            scaled_attention_logits += (mask * -1e9)

            # softmax is normalized on the last axis (seq_len_k) so that the scores
        # add up to 1.
        attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1, name='scaled_dot_prod_attn_weights')  # (..., seq_len_q, seq_len_k)

        output = tf.matmul(attention_weights, v, name='scaled_dot_prod_attn_outs')  # (..., seq_len_q, depth_v)

        return output, attention_weights


MHW_COUNTER = 0
ENC_COUNTER = 0


class MultiHeadAttention(tf.keras.layers.Layer):

    def __init__(self, d_model, num_heads, **kwargs):
        super(MultiHeadAttention, self).__init__(**kwargs)
        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        global MHW_COUNTER
        MHW_COUNTER += 1

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model, name=f"wq_{MHW_COUNTER}")
        self.wk = tf.keras.layers.Dense(d_model, name=f"wk_{MHW_COUNTER}")
        self.wv = tf.keras.layers.Dense(d_model, name=f"wv_{MHW_COUNTER}")

        self.dense = tf.keras.layers.Dense(d_model)

    def split_heads(self, x, batch_size):
        """Split the last dimension into (num_heads, depth).
        Transpose the result such that the shape is (batch_size, num_heads, seq_len, depth)
        """
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def __call__(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)  # (batch_size, seq_len, d_model)
        v = self.wv(v)  # (batch_size, seq_len, d_model)

        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

        # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = scaled_dot_product_attention(
            )(q, k, v, mask)

        scaled_attention = tf.transpose(scaled_attention,
                                        perm=[0, 2, 1, 3])  # (batch_size, seq_len_q, num_heads, depth)

        concat_attention = tf.reshape(scaled_attention,
                                      (batch_size, -1, self.d_model))  # (batch_size, seq_len_q, d_model)

        output = self.dense(concat_attention)  # (batch_size, seq_len_q, d_model)

        return output, attention_weights


def point_wise_feed_forward_network(d_model, dff):
    return tf.keras.Sequential([
        tf.keras.layers.Dense(dff, activation='swish', name='swished_dense'),  # (batch_size, seq_len, dff)
        tf.keras.layers.Dense(d_model, name='ffn_output')  # (batch_size, seq_len, d_model)
    ])


class EncoderLayer(tf.keras.layers.Layer):

    def __init__(self, d_model, num_heads, dff, rate=0.1, **kwargs):
        super(EncoderLayer, self).__init__(**kwargs)

        global MHW_COUNTER
        MHW_COUNTER += 1

        self.mha = MultiHeadAttention(d_model, num_heads)
        # self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.swished_dense = layers.Dense(dff, activation='swish', name=f'swished_dense_{MHW_COUNTER}')
        self.ffn_output = layers.Dense(d_model, name=f'ffn_output_{MHW_COUNTER}')

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def __call__(self, x, training=True, mask=None):
        attn_output, attn_weights = self.mha(x, x, x, mask)  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(x + attn_output)  # (batch_size, input_seq_len, d_model)

        # ffn_output = self.ffn(out1)  # (batch_size, input_seq_len, d_model)

        temp = self.swished_dense(out1)
        ffn_output = self.ffn_output(temp)

        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)  # (batch_size, input_seq_len, d_model)

        return out2, attn_weights


class TransformerEncoder(tf.keras.layers.Layer):

    def __init__(self, num_layers, d_model, num_heads, dff,
                 maximum_position_encoding,
                 rate=0.1, return_weights=False, **kwargs):
        super(TransformerEncoder, self).__init__(**kwargs)

        self.d_model = d_model
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.dff = dff
        self.maximum_position_encoding = maximum_position_encoding
        self.rate = rate
        self.return_weights = return_weights

        #         self.pos_encoding = positional_encoding(self.maximum_position_encoding,
        #                                                 self.d_model)
        #         self.embedding = tf.keras.layers.Dense(self.d_model)
        self.pos_emb = tf.keras.layers.Embedding(input_dim=self.maximum_position_encoding,
                                                 output_dim=self.d_model)

        self.enc_layers = [EncoderLayer(self.d_model, self.num_heads, self.dff, self.rate)
                           for _ in range(self.num_layers)]

        self.dropout = tf.keras.layers.Dropout(self.rate)

        self.attn_weights = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({
            'num_layers': self.num_layers,
            'd_model': self.d_model,
            'num_heads': self.num_heads,
            'dff': self.dff,
            'maximum_position_encoding': self.maximum_position_encoding,
            'dropout': self.dropout,
        })
        return config

    # def call(self, x, training, mask=None):
    def __call__(self, x, training=True, mask=None, *args, **kwargs):

        seq_len = tf.shape(x)[1]

        # adding embedding and position encoding.
        #         x += self.pos_encoding[:, :seq_len, :]
        #         x = self.embedding(x)
        positions = tf.range(start=0, limit=seq_len, delta=1)
        x += self.pos_emb(positions)

        x = self.dropout(x, training=training)
        self.attn_weights = []

        for i in range(self.num_layers):
            x, _w = self.enc_layers[i](x, training, mask)
            self.attn_weights.append(_w)

        if self.return_weights:
            return x, self.attn_weights

        return x  # (batch_size, input_seq_len, d_model)


class Conditionalize(tf.keras.layers.Layer):
    """Mimics the behaviour of cond_rnn of Philipperemy but puts the logic
    of condition in a separate layer so that it becomes easier to use it.

    Example
    --------
    >>> from ai4water.models._tensorflow import Conditionalize
    >>> from tensorflow.keras.layers import Input, LSTM
    >>> i = Input(shape=(10, 3))
    >>> raw_conditions = Input(shape=(14,))
    >>> processed_conds = Conditionalize(32)([raw_conditions, raw_conditions, raw_conditions])
    >>> rnn = LSTM(32)(i, initial_state=[processed_conds, processed_conds])
    """
    def __init__(self, units, max_num_cond=10, **kwargs):
        self.units = units
        super().__init__(**kwargs)

        # single cond
        self.cond_to_init_state_dense_1 = tf.keras.layers.Dense(units=self.units, name="conditional_dense")

        # multi cond
        self.multi_cond_to_init_state_dense = []

        for i in range(max_num_cond):
            self.multi_cond_to_init_state_dense.append(tf.keras.layers.Dense(units=self.units, name=f"conditional_dense{i}"))

        self.multi_cond_p = tf.keras.layers.Dense(1, activation=None, use_bias=True, name="conditional_dense_out")

    @staticmethod
    def _standardize_condition(initial_cond):

        assert len(initial_cond.shape) == 2

        return initial_cond

    def __call__(self, inputs, *args, **kwargs):

        if args or kwargs:
            raise ValueError(f"Unrecognized input arguments\n args: {args} \nkwargs: {kwargs}")

        if inputs.__class__.__name__ == "Tensor":
            inputs = [inputs]

        assert (isinstance(inputs, list) or isinstance(inputs, tuple)) and len(inputs) >= 1, f"{type(inputs)}"

        cond = inputs
        if len(cond) > 1:  # multiple conditions.
            init_state_list = []
            for ii, c in enumerate(cond):
                init_state_list.append(self.multi_cond_to_init_state_dense[ii](self._standardize_condition(c)))
            multi_cond_state = tf.stack(init_state_list, axis=-1)  # -> (?, units, num_conds)
            multi_cond_state = self.multi_cond_p(multi_cond_state)  # -> (?, units, 1)
            cond_state = tf.squeeze(multi_cond_state, axis=-1)  # -> (?, units)
        else:

            cond = self._standardize_condition(cond[0])
            cond_state = self.cond_to_init_state_dense_1(cond)    # -> (?, units)

        return cond_state


class PrivateLayers(object):

    class layers:

        TransformerEncoder = TransformerEncoder
        BasicBlock = BasicBlock
        CONDRNN = ConditionalRNN
        Conditionalize = Conditionalize




import os
from typing import Union
from collections import OrderedDict

import numpy as np
import pandas as pd
from easy_mpl import pie
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

try:
    import shapefile
except ModuleNotFoundError:
    shapefile = None

from .spatial_utils import find_records
from .spatial_utils import plot_shapefile
from .spatial_utils import get_total_area, GifUtil
from .spatial_utils import get_sorted_dict, get_areas_geoms, check_shp_validity


M2ToAcre = 0.0002471     # meter square to Acre
COLORS = ['#CDC0B0', '#00FFFF', '#76EEC6', '#C1CDCD', '#E3CF57', '#EED5B7',
          '#8B7D6B', '#0000FF', '#8A2BE2', '#9C661F', '#FF4040', '#8A360F',
          '#98F5FF', '#FF9912', '#B23AEE', '#9BCD9B', '#8B8B00']


class MakeHRUs(object):
    """
    Distributes a given time series data for HRUs in a catchment according to
    the `hru_definition`. Currently it is supposed that only land use changes
    with time.

    Example:
        >>> import os
        >>> from ai4water.preprocessing.spatial_processing import MakeHRUs
        >>> # shapefile_paths is the path where shapefiles are located. todo
        >>> SubBasin_shp = os.path.join(shapefile_paths, 'sub_basins.shp')
        >>> shapefile_paths = os.path.join(os.getcwd(), 'shapefiles')
        >>> hru_object = MakeHRUs('unique_lu_sub',
        ...     index={2011: {'shapefile': os.path.join(shapefile_paths, 'lu2011.shp'), 'feature': 'NAME'},
        ...     2012: {'shapefile': os.path.join(shapefile_paths, 'lu2012.shp'), 'feature': 'NAME'}},
        ...                  subbasins_shape={'shapefile': SubBasin_shp, 'feature': 'id'}
        ...                 )
        >>> hru_object.call()
    """

    HRU_DEFINITIONS = [
        'unique_sub',
        'unique_lu',
        'unique_soil',
        'unique_slope',
        'unique_lu_sub',
        'unique_lu_soil',
        'unique_lu_slope',
        'unique_soil_sub',
        'unique_soil_slope',
        'unique_slope_sub',
        'unique_lu_soil_sub',
        'unique_lu_soil_slope',
    ]

    def __init__(self,
                 hru_definition:str,
                 index:dict,
                 soil_shape: Union[dict, None] = None,
                 slope_shape: Union[dict, None]=None,
                 subbasins_shape: Union[None, dict] = None,
                 verbosity: int = 1
                 ):
        """
        Arguments:
            hru_definition :
                hru definition. For valid hru_definitions check `MakeHRUs.HRU_DEFINITIONS`
            index :
                dictionary defining shapefiles of landuse which change with time.
                For example in following a land use shapefile for two years is defined.
                All attributes in land use shapefiles must have the feature `NAME`.

                >>> {2011: {'shapefile': os.path.join(shapefile_paths, 'lu2011.shp'), 'feature': 'NAME'},
                ... 2012: {'shapefile': os.path.join(shapefile_paths, 'lu2012.shp'), 'feature': 'NAME'}}

            soil_shape :
                only applicable if `soil` exists in hru_definition.
                All attributes in land use soil.shp must have the feature `NAME`.

                >>> {'shapefile': os.path.join(shapefile_paths, 'soil.shp'), 'feature': 'NAME'}

            slope_shape :
                only applicable if `slope` exists in hru_definition.
                All attributes in slope.shp shapefiles must have the feature `percent`.

                >>> {'shapefile': os.path.join(shapefile_paths, 'slope.shp'), 'feature': 'percent'}

            subbasins_shape :
                only applicable if `sub` exists in hru_definition.
                All attributes in land use shapefiles must have the feature `id`.

                >>> {'shapefile': os.path.join(shapefile_paths, 'subbasins.shp'), 'feature': 'id'}

            verbosity : Determines verbosity.
        """

        self.hru_definition = hru_definition

        assert hru_definition in self.HRU_DEFINITIONS, f"""
        invalid value for hru_definition '{hru_definition}' provided.
        Allowed values are 
        {self.HRU_DEFINITIONS}"""

        self.combinations = hru_definition.split('_')[1:]

        if len(self.combinations)<2:
            if isinstance(index, dict):
                if not all([i.__class__.__name__=='NoneType' for i in index.values()]):
                    assert all([i.__class__.__name__=='NoneType' for i in [soil_shape, slope_shape, subbasins_shape]]), f"""
                    hru consists of only one feature i.e. {self.combinations[0]}. Thus if index is provided then not
                    other shapefile must be given.
                    """

        self.index = index
        self.soil_shape = soil_shape
        self.slope_shape = slope_shape
        self.sub_shape = subbasins_shape
        self.hru_paras = OrderedDict()
        self.hru_geoms = OrderedDict()
        self.all_hrus = []
        self.hru_names = []
        self.verbosity = verbosity

        st, en = list(index.keys())[0], list(index.keys())[-1]
        # initiating yearly dataframes
        self.area = pd.DataFrame(index=pd.date_range(str(st) + '0101', str(en) + '1231', freq='12m'))
        self.curve_no = pd.DataFrame(index=pd.date_range(str(st) + '0101', str(en) + '1231', freq='12m'))
        # distance_to_outlet
        self.dist_to_out = pd.DataFrame(index=pd.date_range(str(st) + '0101', str(en) + '1231',
                                                            freq='12m'))
        # area of HRU as fraction of total catchment
        self.area_frac_cat = pd.DataFrame(index=pd.date_range(str(st) + '0101', str(en) + '1231',
                                                              freq='12m'))

    def call(self, plot_hrus=True):
        """
        Makes the HRUs.
        Arguments:
            plot_hrus : If true, the exact area hrus will be plotted as well.
        """
        for _yr, shp_file in self.index.items():

            _hru_paras, _hru_geoms = self.get_hrus(shp_file, _yr)
            self.hru_paras.update(_hru_paras)

            if plot_hrus:
                self.plot_hrus(year=_yr, _polygon_dict=self.hru_geoms, nrows=3, ncols=4,
                               bbox=self.slope_shape, annotate=False, save=False,
                               name=self.hru_definition)
        return

    def get_hrus(self,
                 idx_shp,  # shapefile whose area distribution changes with time e.g land use
                 year
                 ):
        """
        lu_path: path of landuse shapefile
        :return:

        """

        if self.verbosity > 0:
            print('Checking validity of landuse shapefile')

        hru_paras = OrderedDict()
        a = 0

        if len(self.combinations) ==1:
            shp_name = self.combinations[0]
            if idx_shp is None:
                shp_file = getattr(self, f'{shp_name}_shape')['shapefile']
                feature = getattr(self, f'{shp_name}_shape')['feature']
            else:
                shp_file = idx_shp['shapefile']
                feature = idx_shp['feature']

            shp_reader = shapefile.Reader(shp_file)
            _, shp_geom_list = get_areas_geoms(shp_reader)
            if 'sub' not in self.hru_definition:
                shp_geom_list = check_shp_validity(shp_geom_list, len(shp_reader.shapes()), name=shp_name,
                                                   verbosity=self.verbosity)

            self.tot_cat_area = get_total_area(shp_geom_list)

            for shp in range(len(shp_reader.shapes())):
                code = f'{str(year)}_{shp_name}_{find_records(shp_file, feature, shp)}'
                hru_paras[code] = {'yearless_key': code[4:]}
                intersection = shp_geom_list[shp]
                self.hru_geoms[code] = [intersection, shp_geom_list[shp]]
                self._foo(code, intersection)

        if len(self.combinations) == 2:
            first_shp_name = self.combinations[0]
            second_shp_name = self.combinations[1]

            if idx_shp is None:
                first_shp_file = getattr(self, f'{first_shp_name}_shape')['shapefile']
                first_feature = getattr(self, f'{first_shp_name}_shape')['feature']
            else:
                first_shp_file = idx_shp['shapefile']
                first_feature = idx_shp['feature']

            second_shp_file = getattr(self, f'{second_shp_name}_shape')['shapefile']
            second_feature = getattr(self, f'{second_shp_name}_shape')['feature']

            first_shp_reader = shapefile.Reader(first_shp_file)
            second_shp_reader = shapefile.Reader(second_shp_file)

            _, first_shp_geom_list = get_areas_geoms(first_shp_reader)
            _, second_shp_geom_list = get_areas_geoms(second_shp_reader)

            if 'sub' not in self.hru_definition:
                second_shp_geom_list = check_shp_validity(second_shp_geom_list, len(second_shp_reader.shapes()),
                                                          name=second_shp_name, verbosity=self.verbosity)
                self.tot_cat_area = get_total_area(first_shp_geom_list)
            else:
                self.tot_cat_area = get_total_area(second_shp_geom_list)

            for j in range(len(second_shp_reader.shapes())):
                for lu in range(len(first_shp_reader.shapes())):
                    a += 1

                    intersection = second_shp_geom_list[j].intersection(first_shp_geom_list[lu])

                    lu_code = find_records(first_shp_file, first_feature, lu)
                    sub_code = find_records(second_shp_file, second_feature, j)
                    sub_code = f'_{second_shp_name}_' + str(sub_code)
                    code = str(year) + sub_code + f'_{first_shp_name}_' + lu_code #, lu_code

                    self.hru_geoms[code] = [intersection, second_shp_geom_list[j], first_shp_geom_list[lu]]

                    hru_paras[code] = {'yearless_key': code[4:]}
                    self._foo(code, intersection)

        if len(self.combinations) == 3:
            first_shp_name = self.combinations[0]
            second_shp_name = self.combinations[1]
            third_shp_name = self.combinations[2]

            if idx_shp is None:
                first_shp_file = getattr(self, f'{first_shp_name}_shape')['shapefile']
                first_feature = getattr(self, f'{first_shp_name}_shape')['feature']
            else:
                first_shp_file = idx_shp['shapefile']
                first_feature = idx_shp['feature']

            second_shp_file = getattr(self, f'{second_shp_name}_shape')['shapefile']
            second_feature = getattr(self, f'{second_shp_name}_shape')['feature']
            third_shp_file = getattr(self, f'{third_shp_name}_shape')['shapefile']
            third_feature = getattr(self, f'{third_shp_name}_shape')['feature']

            first_shp_reader = shapefile.Reader(first_shp_file)
            second_shp_reader = shapefile.Reader(second_shp_file)
            third_shp_reader = shapefile.Reader(third_shp_file)

            _, first_shp_geom_list = get_areas_geoms(first_shp_reader)
            _, second_shp_geom_list = get_areas_geoms(second_shp_reader)
            _, third_shp_geom_list = get_areas_geoms(third_shp_reader)

            if 'sub' not in self.hru_definition:  # todo
                second_shp_geom_list = check_shp_validity(second_shp_geom_list, len(second_shp_reader.shapes()),
                                                          name=second_shp_name,
                                                          verbosity=self.verbosity)
                self.tot_cat_area = get_total_area(first_shp_geom_list)
            else:
                self.tot_cat_area = get_total_area(second_shp_geom_list)

            for s in range(len(third_shp_reader.shapes())):
                for j in range(len(second_shp_reader.shapes())):
                    for lu in range(len(first_shp_reader.shapes())):

                        intersection = second_shp_geom_list[j].intersection(first_shp_geom_list[lu])

                        sub = third_shp_geom_list[s]
                        intersection = sub.intersection(intersection)

                        sub_code = f'_{third_shp_name}_' + str(find_records(third_shp_file, third_feature, s))
                        lu_code = find_records(first_shp_file, first_feature, lu)
                        soil_code = find_records(second_shp_file, second_feature, j)
                        code = str(year) + sub_code + f'_{second_shp_name}_' + str(soil_code) + f'_{first_shp_name}_' + lu_code

                        self.hru_geoms[code] = [intersection, second_shp_geom_list[j], first_shp_geom_list[lu]]

                        hru_paras[code] = {'yearless_key': code[4:]}
                        self._foo(code, intersection)

        # if sum(LuAreaListAcre) > sum(SubAreaListAcre):
        #     print('Land use area is bigger than subbasin area')
        #     IntArea_ = [[None] * no_of_subs for _ in range(no_of_lus)]
        #     IntLuArea = [None] * no_of_lus
        #     for lu in range(no_of_lus):
        #         # Int = 0
        #         IntArea = 0
        #         for sub_ind in range(no_of_subs):
        #             Int = LuShpGeomListNew[lu].intersection(SubShpGeomList[sub_ind])
        #             code, lu_code = get_code(year=year, lu_feature=LuShpGeomListNew[lu], lu_feat_ind=lu,
        #             sub_feat=SubShpGeomList[sub_ind], sub_feat_ind=sub_ind, _hru_def=_hru_def)
        #             IntArea += Int.area * M2ToAcre
        #             anarea = Int.area * M2ToAcre
        #             IntArea_[lu][sub_ind] = anarea
        #         print('area of {0} is reduced from {1:10.3f} acres to {2:10.3f}'
        #               .format(lu_code, LuAreaListAcre[lu], IntArea))  # lu_code/j
        #         IntLuArea[lu] = IntArea
        #     print('New area of all landuses is  {}'.format(sum(IntLuArea)))
        # else:
        #     print('Land use area is equal to subbasin area')

        self.hru_names = list(set(self.hru_names))

        return hru_paras, self.hru_geoms

    def _foo(self, code, intersection):
        hru_name = code[5:]
        year = code[0:4]
        row_index = pd.to_datetime(year + '0131', format='%Y%m%d', errors='ignore')
        self.hru_names.append(hru_name)
        self.all_hrus.append(code)
        anarea = intersection.area * M2ToAcre

        # saving value of area for currnet HRU and for for current year in dictionary
        self.area.loc[row_index, hru_name] = anarea
        self.area_frac_cat.loc[row_index, hru_name] = anarea / self.tot_cat_area
        return

    def plot_hrus(self, year, bbox, _polygon_dict, annotate=False, nrows=3,
                  ncols=4, save=False, name='',
                  annotate_missing_hru=False):

        polygon_dict = OrderedDict()
        for k, v in _polygon_dict.items():
            if str(year) in k[0:4]:
                polygon_dict[k] = v

        # sorting dictionary based on keys so that it puts same HRU at same place for each year
        polygon_dict = get_sorted_dict(polygon_dict)

        figure, axs = plt.subplots(nrows, ncols)

        if isinstance(bbox, str):
            r = shapefile.Reader(bbox)
            bbox = r.bbox
            r.close()

        figure.set_figwidth(27)
        figure.set_figheight(12)
        axis_l = [item for sublist in list(axs) for item in sublist]
        # max_bbox = get_bbox_with_max_area(_polygon_dict)

        i = 0
        for key, axis in zip(polygon_dict, axis_l):
            i += 1
            ob = polygon_dict[key][0]
            # text0 = key.split('_')[4]+' in '+key.split('_')[1]  +' '+ key.split('_')[2]
            if ob.type == 'MultiPolygon':
                anfang_x = [None] * len(ob)
                for s_ob in range(len(ob)):
                    ob_ = ob[s_ob]
                    x, y = ob_.exterior.xy
                    axis.plot(x, y, color=np.random.rand(3, ), alpha=0.7, linewidth=3, solid_capstyle='round', zorder=2)
                    axis.get_yaxis().set_visible(False)
                    axis.get_xaxis().set_visible(False)
                    if bbox is not None:
                        axis.set_ylim([bbox[1], bbox[3]])
                        axis.set_xlim([bbox[0], bbox[2]])
                    anfang_x[s_ob] = x[0]
                if annotate:
                    axis.annotate(key[3:], xy=(0.2, 0.1), xycoords='axes fraction', fontsize=18)

            elif ob.type == 'Polygon':
                x, y = polygon_dict[key][0].exterior.xy
                axis.plot(x, y, color=np.random.rand(3, ), alpha=0.7, linewidth=3, solid_capstyle='round', zorder=2)
                axis.get_yaxis().set_visible(False)
                axis.get_xaxis().set_visible(False)
                if annotate:
                    axis.annotate(key[3:], xy=(0.2, 0.1), xycoords='axes fraction', fontsize=18)
                if bbox is not None:
                    axis.set_ylim([bbox[1], bbox[3]])
                    axis.set_xlim([bbox[0], bbox[2]])
                # axis.text(x[0], np.mean(y), text0, color='red', fontsize='12')

            else:  # for empty cases
                x, y = np.arange(0, 10), np.arange(0, 10)
                axis.plot(x, y, color='w')
                text1 = 'no ' + key.split('_')[1] + ' in sub-basin ' + key.split('_')[2]
                if annotate_missing_hru:
                    axis.text(x[0], np.mean(y), text1, color='red', fontsize=16)
                axis.get_yaxis().set_visible(False)
                axis.get_xaxis().set_visible(False)

        figure.suptitle('HRUs for year {}'.format(year), fontsize=22)
        # plt.title('HRUs for year {}'.format(year), fontsize=22)
        if save:
            name = 'hrus_{}.png'.format(year) if name is None else name + str(year)
            plt.savefig('plots/' + name)
        plt.show()

    def plot_as_ts(self, save=False, name=None, show=True, **kwargs):
        """hru_object.plot_as_ts(save=True, min_xticks=3, max_xticks=4"""

        figsize = kwargs.get('figsize', (12, 6))
        bbox_inches = kwargs.get('bbox_inches', 'tight')
        tick_fs = kwargs.get('tick_fs', 14)
        axis_label_fs = kwargs.get('axis_label_fs', 18)
        bbox_to_anchor = kwargs.get('bbox_to_anchor', (1.1, 0.99))
        leg_fs = kwargs.get('leg_fs', 14)
        markerscale = kwargs.get('markerscale', 2)
        style=  kwargs.get('style', '-s')
        title = kwargs.get('title', False)
        min_xticks = kwargs.get('min_xticks', None)
        max_xticks = kwargs.get('max_xticks', None)

        plt.close('all')
        _, axis = plt.subplots(figsize=figsize)
        for area in self.area:
            axis.plot(self.area[area], style, label=area)
        axis.legend(fontsize=leg_fs, markerscale=markerscale, bbox_to_anchor=bbox_to_anchor)
        axis.set_xlabel('Time', fontsize=axis_label_fs)
        axis.set_ylabel('Area (Acres)', fontsize=axis_label_fs)
        axis.tick_params(axis="x", which='major', labelsize=tick_fs)
        axis.tick_params(axis="y", which='major', labelsize=tick_fs)

        if min_xticks is not None:
            assert isinstance(min_xticks, int)
            assert isinstance(max_xticks, int)
            loc = mdates.AutoDateLocator(minticks=4, maxticks=6)
            axis.xaxis.set_major_locator(loc)
            fmt = mdates.AutoDateFormatter(loc)
            axis.xaxis.set_major_formatter(fmt)
        if title:
            plt.suptitle('Variation of Area (acre) of HRUs with time')
        if save:
            if name is None:
                name = self.hru_definition
            plt.savefig(f'{name}_hru_as_ts.png', dpi=300, bbox_inches=bbox_inches)

        if show:
            plt.show()

        return

    def plot_hru_evolution(self, hru_name, make_gif=False):
        """
        plots how the hru evolved during the years
        :param hru_name: str, name of hru to be plotted
        :param make_gif: bool, if True, a gif file will be created from evolution plots
        :return: plots the hru.
        """
        for yr in self.index.keys():
            y = str(yr)[2:] + '_'
            hru_name_year = y + hru_name  # hru name with year
            self.plot_hru(hru_name_year, self.soil_shape)

        if make_gif:
            gif = GifUtil(folder=os.path.join(os.getcwd(), 'plots'), contains=hru_name)
            gif.make_gif()
            gif.remove_images()
        return

    def make_gif(self):
        gif = GifUtil(initials=self.hru_definition, folder=os.path.join(os.getcwd(), 'plots'))
        gif.make_gif()
        gif.remove_images()
        return

    def plot(self, what, index=None, show_all_together=True):
        assert what in ['landuse', 'soil', 'subbasins', 'slope']
        if what == 'landuse':
            assert index
            shp_file = self.index[index]
        else:
            shp_file = getattr(self, f'{what}_shape')

        plot_shapefile(shp_file, show_all_together)
        return

    def plot_hru(self, hru_name, bbox=None, save=False):
        """
        plot only one hru from `hru_geoms`. The value of each key in hru_geoms is a list with three shapes
        :usage
          self.plot_an_hru(self.hru_names[0], bbox=True)
        """
        shape_list = self.hru_geoms[hru_name]

        figure, (axis_list) = plt.subplots(3)
        figure.set_figheight(14)

        if bbox:
            r = shapefile.Reader(bbox)
            bbox = r.bbox

        i = 0
        leg = hru_name
        for axs, ob in zip(axis_list, shape_list):
            i +=1

            if i==2: leg = hru_name.split('_')[1:2]
            elif i==3: leg = hru_name.split('_')[-1]

            if ob.type == 'MultiPolygon':
                for s_ob in range(len(ob)):
                    ob_ = ob[s_ob]
                    x, y = ob_.exterior.xy
                    axs.plot(x, y, color=COLORS[i], alpha=0.7, linewidth=3, solid_capstyle='round', zorder=2)
            elif ob.type == 'Polygon':
                x, y = ob.exterior.xy
                axs.plot(x, y, color=COLORS[i], alpha=0.7, linewidth=3, solid_capstyle='round', zorder=2)
            axs.set_title(leg, fontsize=14)

            if bbox:
                axs.set_ylim([bbox[1], bbox[3]])
                axs.set_xlim([bbox[0], bbox[2]])

        if save:
            plt.savefig('plots/' + hru_name + '.png')
        plt.show()
        return

    def draw_pie(self,
                 year:int,
                 n_merge:int=0,
                 title:bool=False,
                 save:bool=False,
                 name:str=None,
                 show:bool = True,
                 **kwargs):
        """
        todo draw nested pie chart for all years
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.pie.html
        Draws a pie chart showing relative area of HRUs for a particular year.
        Since the hrus can change with time, selecting one year is based on supposition
        that area of hrus remain constant during the whole year.
        Arguments:
            year : int, the year for which area of hrus will be used.
            n_merge : number of hrus to merge
            title :
            save :
            name :
            show :
            kwargs : Following keyword arguments are allowed
                shadow
                strartangle
                autopct
                textprops
        """
        shadow = kwargs.get('shadow', True)
        startangle = kwargs.get('startangle', 90)
        autopct = kwargs.get('autopct', '%1.1f%%')
        textprops = kwargs.get('textprops', {})

        idx = str(year) + '-01-31'
        area_unsort = self.area.loc[idx]
        area = area_unsort.sort_values()
        merged = area[area.index[0]:area.index[n_merge-1]]
        rest = area[area.index[n_merge]:]

        if n_merge==0:
            assert len(merged) == len(area)
            merged_vals= []
            merged_labels =  []
        else:
            merged_vals = [sum(merged.values)]
            merged_labels = ['{} HRUs'.format(n_merge)]

        vals = list(rest.values) + merged_vals
        labels = list(rest.keys()) + merged_labels

        explode = [0 for _ in range(len(vals))]
        explode[-1] = 0.1

        labels_n = []
        for l in labels:
            labels_n.append(l.replace('lu_', ''))

        if title:
            title = 'Areas of HRUs for year {}'.format(year)

        if name is None: name = self.hru_definition
        name = f'{len(self.hru_names)}hrus_for_{year}_{name}.png'

        return pie(fractions=vals,
                   labels=labels_n,
                   explode=tuple(explode),
                   autopct=autopct, shadow=shadow, startangle=startangle, textprops=textprops,
                   title=title, name=name, save=save, show=show)



# -*- coding: utf-8 -*-
# Don't know which rights should be reserved  Ather Abbas
from setuptools import setup


import os
fpath = os.path.join(os.getcwd(), "README.md")
if os.path.exists(fpath):
    with open(fpath, "r") as fd:
        long_desc = fd.read()
else:
    long_desc = "https://github.com/AtrCheema/AI4Water"

min_requirements = [
    'numpy<=1.19.5',
    'scikit-learn',
    'pandas',
    'matplotlib',
    'scikit-optimize',
    'joblib',
    'requests',
    'plotly',
    'easy_mpl>=0.20.3',
    'SeqMetrics>=1.3.2'
    ]

extra_requires = [
'tensorflow', # only if you want to use tensorflow-based models, >=1.15, 2.4 having trouble with see-rnn
'scikit-optimize',  # only if you want to use file hyper_opt.py for hyper-parameter optimization

'h5py<2.11.0', # only if you want to save batches
'xgboost',
'lightgbm',
'catboost',
'tpot',
# spatial processing
'imageio',
# shapely manually download the wheel file and install
'pyshp',

'optuna',
'hyperopt',

# for reading data
'netCDF4',
 'xarray',

# for jsonizing
'wrapt',

# eda
'seaborn'
]

tf_requires = ['h5py<2.11.0', 'numpy<=1.19.5', 'easy_mpl', 'tensorflow', 'pandas',
               'matplotlib', 'scikit-learn', 'SeqMetrics>=1.3.2']

tf_hpo_requires = ['h5py<2.11.0', 'numpy<=1.19.5', 'easy_mpl', 'tensorflow', 'pandas',
                   'matplotlib', 'scikit-learn', 'hyperopt', 'scikit-optimize', 'optuna',
                   'SeqMetrics>=1.3.2']

torch_requires = ['h5py', 'numpy', 'easy_mpl>=0.20.3',  'pytorch', 'pandas',
                  'matplotlib', 'scikit-learn', 'SeqMetrics>=1.3.2']

torch_hpo_requires = ['h5py', 'numpy', 'easy_mpl>=0.20.3',  'pytorch', 'pandas',
                  'matplotlib', 'scikit-learn', 'hyperopt', 'scikit-optimize', 'optuna',
                      'SeqMetrics>=1.3.2']

ml_requires = ['numpy', 'matplotlib', 'pandas', 'scikit-learn', 'xgboost', 'catboost',
               'lightgbm', 'easy_mpl>=0.20.3', 'SeqMetrics>=1.3.2']

ml_hpo_requires = ['numpy', 'matplotlib', 'pandas', 'scikit-learn', 'xgboost', 'catboost',
               'lightgbm', 'easy_mpl>=0.20.3', 'hyperopt', 'scikit-optimize', 'optuna',
                   'SeqMetrics>=1.3.2']

hpo_requirements = ['optuna', 'hyperopt', 'scikit-optimize', 'SeqMetrics>=1.3.2']

post_process_requirements = ['lime', 'shap', 'SeqMetrics>=1.3.2']

exp_requirements = ['catboost', 'lightgbm', 'xgboost',
                    'tpot',
                    'optuna', 'hyperopt', 'scikit-optimize',
                    'h5py<2.11.0', 'SeqMetrics>=1.3.2'
                    ]
pre_prcess_requirements = ['netCDF4', 'xarray', 'imageio', 'pyshp', 'SeqMetrics>=1.3.2']

eda_requires = ['seaborn', 'scikit-learn', 'easy_mpl>=0.20.3', 'SeqMetrics>=1.3.2']

all_requirements = min_requirements + extra_requires

setup(

    name='AI4Water',

    version="1.0b4",

    description='Platform for developing data driven based models for sequential/tabular data',
    long_description=long_desc,
    long_description_content_type="text/markdown",

    url='https://github.com/AtrCheema/AI4Water',

    author='Ather Abbas',
    author_email='ather_abbas786@yahoo.com',

    package_data={'ai4water/datasets': ['arg_busan.csv']},
    include_package_data=True,

    classifiers=[
        'Development Status :: 4 - Beta',

        'Natural Language :: English',

        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    packages=['ai4water',
              'ai4water/models',
              'ai4water/models/tensorflow',
              'ai4water/models/torch',
              'ai4water/hyperopt',
              'ai4water/utils',
              'ai4water/preprocessing',
              'ai4water/preprocessing/transformations',
              'ai4water/preprocessing/dataset',
              'ai4water/postprocessing/',
              'ai4water/postprocessing/SeqMetrics',
              'ai4water/postprocessing/explain',
              'ai4water/postprocessing/interpret',
              'ai4water/postprocessing/visualize',
              'ai4water/datasets',
              'ai4water/et',
              'ai4water/experiments',
              'ai4water/eda'
              ],

    install_requires=min_requirements,

    extras_require={
        'all': extra_requires,
        'hpo': hpo_requirements,
        'post_process': post_process_requirements,
        'exp': exp_requirements,
        'eda': eda_requires,
        'tf': tf_requires,
        'torch': torch_requires,
        'tf_hpo': tf_hpo_requires,
        'torch_hpo_requires': torch_hpo_requires,
        'ml': ml_requires,
        'ml_hpo': ml_hpo_requires,
    }
)



import os
import itertools
from types import FunctionType
from collections import OrderedDict

import numpy as np
import matplotlib.pyplot as plt
from scipy.special import xlogy

try:
    import plotly.graph_objects as go
except ModuleNotFoundError:
    go = None


def take(st, en, d):
    keys = list(d.keys())[st:en]
    values = list(d.values())[st:en]

    return {k: v for k, v in zip(keys, values)}


def plot_metrics(
        metrics: dict,
        ranges: tuple = ((0.0, 1.0), (1.0, 10), (10, 1000)),
        exclude: list = None,
        plot_type: str = 'bar',
        max_metrics_per_fig: int = 15,
        show: bool = True,
        save: bool = False,
        save_path: str = None,
        **kwargs):
    """
    Plots the metrics given as dictionary as radial or bar plot between specified ranges.

    Arguments:
        metrics:
            dictionary whose keys are names are erros and values are error values.
        ranges:
            tuple of tuples defining range of errors to plot in one plot
        exclude:
            List of metrics to be excluded from plotting.
        max_metrics_per_fig:
            maximum number of metrics to show in one figure.
        plot_type:
            either of `radial` or `bar`.
        show : If, then figure will be shown/drawn
        save:
            if True, the figure will be saved.
        save_path:
            if given, the figure will the saved at this location.
        kwargs:
            keyword arguments for plotting

    Examples:
        >>> import numpy as np
        >>> from ai4water.postprocessing.SeqMetrics import RegressionMetrics
        >>> from ai4water.postprocessing.SeqMetrics import plot_metrics
        >>> t = np.random.random((20, 1))
        >>> p = np.random.random((20, 1))
        >>> er = RegressionMetrics(t, p)
        >>> all_errors = er.calculate_all()
        >>> plot_metrics(all_errors, plot_type='bar', max_metrics_per_fig=50)
        >>># or draw the radial plot
        >>> plot_metrics(all_errors, plot_type='radial', max_metrics_per_fig=50)
    ```
    """
    for idx, rng in enumerate(ranges):
        assert rng[1] > rng[0], f'For range {idx}, second value: {rng[1]} is not greater than first value: {rng[0]}. '
        assert len(rng) == 2, f"Range number {idx} has length {len(rng)}. It must be a tuple of length 2."

    if exclude is None:
        exclude = []

    _metrics = metrics.copy()
    for k in metrics.keys():
        if k in exclude:
            _metrics.pop(k)

    assert plot_type in ['bar', 'radial'], f'plot_type must be either `bar` or `radial`.'

    for _range in ranges:
        plot_metrics_between(
            _metrics,
            *_range,
            plot_type=plot_type,
            max_metrics_per_fig=max_metrics_per_fig,
            show=show,
            save=save,
            save_path=save_path, **kwargs)
    return


def plot_metrics_between(
        errors: dict,
        lower: int,
        upper: int,
        plot_type: str = 'bar',
        max_metrics_per_fig: int = 15,
        save=False,
        show=True,
        save_path=None,
        **kwargs):
    zero_to_one = {}
    for k, v in errors.items():
        if v is not None:
            if lower < v < upper:
                zero_to_one[k] = v
    st = 0
    n = len(zero_to_one)
    for i in np.array(np.linspace(0, n, int(n/max_metrics_per_fig)+1),
                      dtype=np.int32):
        if i == 0:
            pass
        else:
            en = i
            d = take(st, en, zero_to_one)
            if plot_type == 'radial':
                plot_radial(d, lower, upper, save=save, show=show, save_path=save_path, **kwargs)
            else:
                plot_circular_bar(d, save=save, show=show, save_path=save_path, **kwargs)
            st = i
    return


def plot_radial(errors: dict, low: int, up: int, save=True, save_path=None, **kwargs):
    """Plots all the errors in errors dictionary. low and up are used to draw the limits of radial plot."""
    if go is None:
        print("can not plot radial plot because plotly is not installed.")
        return

    fill = kwargs.get('fill', None)
    fillcolor = kwargs.get('fillcolor', None)
    line = kwargs.get('line', None)
    marker = kwargs.get('marker', None)

    OrderedDict(sorted(errors.items(), key=lambda kv: kv[1]))

    lower = round(np.min(list(errors.values())), 4)
    upper = round(np.max(list(errors.values())), 4)

    fig = go.Figure()
    categories = list(errors.keys())

    fig.add_trace(go.Scatterpolar(
        r=list(errors.values()),
        theta=categories,  # angular coordinates
        fill=fill,
        fillcolor=fillcolor,
        line=line,
        marker=marker,
        name='errors'
    ))

    fig.update_layout(
        title_text=f"Errors from {lower} to {upper}",
        polar=dict(
            radialaxis=dict(
                visible=True,
                range=[low, up]
            )),
        showlegend=False
    )

    fig.show()
    if save:
        fname = f"radial_errors_from_{lower}_to_{upper}.png"
        if save_path is not None:
            fname = os.path.join(save_path, fname)
        fig.write_image(fname)
    return


def plot_circular_bar(
        metrics: dict,
        show=False,
        save: bool = True,
        save_path: str = '',
        **kwargs):
    """
    modified after https://www.python-graph-gallery.com/circular-barplot-basic
    :param metrics:
    :param show:
    :param save:
    :param save_path:
    :param kwargs:
        figsize:
        linewidth:
        edgecolor:
        color:
    :return:
    """

    # initialize the figure
    plt.close('all')
    plt.figure(figsize=kwargs.get('figsize', (8, 12)))
    ax = plt.subplot(111, polar=True)
    plt.axis('off')

    # Set the coordinates limits
    # upperLimit = 100
    lower_limit = 30
    value = np.array(list(metrics.values()))

    lower = round(np.min(list(metrics.values())), 4)
    upper = round(np.max(list(metrics.values())), 4)

    # Compute max and min in the dataset
    _max = max(value)  # df['Value'].max()

    # Let's compute heights: they are a conversion of each item value in those new coordinates
    # In our example, 0 in the dataset will be converted to the lowerLimit (10)
    # The maximum will be converted to the upperLimit (100)
    slope = (_max - lower_limit) / _max
    heights = slope * value + lower_limit

    # Compute the width of each bar. In total we have 2*Pi = 360°
    width = 2 * np.pi / len(metrics)

    # Compute the angle each bar is centered on:
    indexes = list(range(1, len(metrics) + 1))
    angles = [element * width for element in indexes]

    # Draw bars
    bars = ax.bar(
        x=angles,
        height=heights,
        width=width,
        bottom=lower_limit,
        linewidth=kwargs.get('linewidth', 2),
        edgecolor=kwargs.get('edgecolor', "white"),
        color=kwargs.get('color', "#61a4b2"),
    )

    # little space between the bar and the label
    label_padding = 4

    metric_names = {
        'r2': "$R^2$",
        'r2_mod': "$R^2$ mod",
        'adjusted_r2': 'adjusted $R^2$',
        # 'nse': "NSE"
    }

    # Add labels
    for bar, angle, label1, label2 in zip(bars, angles, metrics.keys(), metrics.values()):

        label1 = metric_names.get(label1, label1)
        label = f'{label1} {round(label2, 4)}'

        # Labels are rotated. Rotation must be specified in degrees :(
        rotation = np.rad2deg(angle)

        # Flip some labels upside down
        if angle >= np.pi / 2 and angle < 3 * np.pi / 2:
            alignment = "right"
            rotation = rotation + 180
        else:
            alignment = "left"

        # Finally add the labels
        ax.text(
            x=angle,
            y=lower_limit + bar.get_height() + label_padding,
            s=label,
            ha=alignment,
            va='center',
            rotation=rotation,
            rotation_mode="anchor")

    if save:
        fname = f"{len(metrics)}_bar_errors_from_{lower}_to_{upper}.png"
        if save_path is not None:
            fname = os.path.join(save_path, fname)
        plt.savefig(fname, dpi=400, bbox_inches='tight')
    if show:
        plt.show()

    return


def plot1d(true, predicted, save=True, name="plot", show=False):
    _, axis = plt.subplots()

    axis.plot(np.arange(len(true)), true, label="True")
    axis.plot(np.arange(len(predicted)), predicted, label="Predicted")
    axis.legend(loc="best")

    if save:
        plt.savefig(name, dpi=300, bbox_inches='tight')
    if show:
        plt.show()

    plt.close('all')
    return


def _foo(denominator, numerator):
    nonzero_numerator = numerator != 0
    nonzero_denominator = denominator != 0
    valid_score = nonzero_numerator & nonzero_denominator
    output_scores = np.ones(1)

    output_scores[valid_score] = 1 - (numerator[valid_score] /
                                      denominator[valid_score])
    output_scores[nonzero_numerator & ~nonzero_denominator] = 0.
    return output_scores


def _mean_tweedie_deviance(y_true, y_pred, power=0, weights=None):
    # copying from
    # https://github.com/scikit-learn/scikit-learn/blob/95d4f0841d57e8b5f6b2a570312e9d832e69debc/sklearn/metrics/_regression.py#L659

    message = ("Mean Tweedie deviance error with power={} can only be used on "
               .format(power))
    if power < 0:
        # 'Extreme stable', y_true any real number, y_pred > 0
        if (y_pred <= 0).any():
            raise ValueError(message + "strictly positive y_pred.")
        dev = 2 * (np.power(np.maximum(y_true, 0), 2 - power)
                   / ((1 - power) * (2 - power))
                   - y_true * np.power(y_pred, 1 - power) / (1 - power)
                   + np.power(y_pred, 2 - power) / (2 - power))
    elif power == 0:
        # Normal distribution, y_true and y_pred any real number
        dev = (y_true - y_pred) ** 2
    elif power < 1:
        raise ValueError("Tweedie deviance is only defined for power<=0 and "
                         "power>=1.")
    elif power == 1:
        # Poisson distribution, y_true >= 0, y_pred > 0
        if (y_true < 0).any() or (y_pred <= 0).any():
            raise ValueError(message + "non-negative y_true and strictly "
                                       "positive y_pred.")
        dev = 2 * (xlogy(y_true, y_true / y_pred) - y_true + y_pred)
    elif power == 2:
        # Gamma distribution, y_true and y_pred > 0
        if (y_true <= 0).any() or (y_pred <= 0).any():
            raise ValueError(message + "strictly positive y_true and y_pred.")
        dev = 2 * (np.log(y_pred / y_true) + y_true / y_pred - 1)
    else:
        if power < 2:
            # 1 < p < 2 is Compound Poisson, y_true >= 0, y_pred > 0
            if (y_true < 0).any() or (y_pred <= 0).any():
                raise ValueError(message + "non-negative y_true and strictly "
                                           "positive y_pred.")
        else:
            if (y_true <= 0).any() or (y_pred <= 0).any():
                raise ValueError(message + "strictly positive y_true and "
                                           "y_pred.")

        dev = 2 * (np.power(y_true, 2 - power) / ((1 - power) * (2 - power))
                   - y_true * np.power(y_pred, 1 - power) / (1 - power)
                   + np.power(y_pred, 2 - power) / (2 - power))

    return float(np.average(dev, weights=weights))


def _geometric_mean(a, axis=0, dtype=None):
    """ Geometric mean """
    if not isinstance(a, np.ndarray):  # if not an ndarray object attempt to convert it
        log_a = np.log(np.array(a, dtype=dtype))
    elif dtype:  # Must change the default dtype allowing array type
        if isinstance(a, np.ma.MaskedArray):
            log_a = np.log(np.ma.asarray(a, dtype=dtype))
        else:
            log_a = np.log(np.asarray(a, dtype=dtype))
    else:
        log_a = np.log(a)
    return float(np.exp(log_a.mean(axis=axis)))


def listMethods(cls):
    return set(x for x, y in cls.__dict__.items()
               if isinstance(y, (FunctionType, classmethod, staticmethod)))


def listParentMethods(cls):
    return set(itertools.chain.from_iterable(
        listMethods(c).union(listParentMethods(c)) for c in cls.__bases__))


def list_subclass_methods(cls, is_narrow, ignore_underscore=True, additional_ignores=None):
    """Finds all methods of a child class"""
    methods = listMethods(cls)

    if is_narrow:
        parent_methods = listParentMethods(cls)
        methods = set(cls for cls in methods if not (cls in parent_methods))
    
    if additional_ignores is not None:
        methods = methods - set(additional_ignores)

    if ignore_underscore:
        methods = set(cls for cls in methods if not cls.startswith('_'))

    return methods

def msg(module, module_type="class"):
    return f"""
    {module} {module_type} has been moved to its own repository called SeqMetrics
    Please install it using 'pip install SeqMetrics' and then import {module}
    as
    from SeqMetrics import {module}
    """


import os
import json
from itertools import islice
from pickle import PicklingError
from collections import OrderedDict

import numpy as np
import pandas as pd
import matplotlib as mpl
from easy_mpl import plot
import matplotlib.pyplot as plt

try:
    from hyperopt import space_eval, hp
    from hyperopt.base import miscs_to_idxs_vals  # todo main_plot_1D_attachment
except ImportError:
    space_eval, hp = None, None
    miscs_to_idxs_vals = None

try:
    from skopt.plots import plot_evaluations, plot_objective
    import skopt
    from skopt.space.space import Space, Dimension
except ImportError:
    plot_evaluations, plot_objective = None, None
    Space = None
    Dimension = None

try:
    from skopt.utils import dump
except ImportError:
    dump = None


from ai4water.utils.utils import Jsonize, clear_weights
from ._space import Categorical, Real, Integer


def is_choice(space):
    """checks if an hp.space is hp.choice or not"""
    if 'switch' in space.name:
        return True
    return False


def x_iter_for_tpe(trials, param_space: dict, as_list=True):

    assert isinstance(param_space, dict)

    x_iters = []  # todo, remove x_iters, it is just values of iterations
    iterations = {}
    for idx, t in enumerate(trials.trials):

        vals = t['misc']['vals']
        y = t['result']['loss']

        _iter = get_one_tpe_x_iter(vals, param_space)
        iterations[f'{round(y, 5)}_{idx}'] = _iter

        x_iters.append(_iter)

    if as_list:
        return [list(d.values()) for d in x_iters]
    return iterations


def get_one_tpe_x_iter(tpe_vals, param_space: dict, sort=True):

    x_iter = {}
    for para, para_val in tpe_vals.items():
        if is_choice(param_space[para]):
            hp_assign = {para: para_val[0]}
            cval = space_eval(param_space[para], hp_assign)
            x_iter[para] = cval
        else:
            x_iter[para] = para_val[0]

    if sort:
        x_iter = sort_x_iters(x_iter, list(param_space.keys()))

    return x_iter


def sort_x_iters(x_iter: dict, original_order: list):
    # the values in x_iter may not be sorted as the parameters provided in original order

    new_x_iter = {}
    for s in original_order:
        new_x_iter[s] = x_iter[s]

    return new_x_iter


def skopt_space_from_hp_spaces(hp_space: dict) -> list:
    """given a dictionary of hp spaces where keys are names, this function
    converts it into skopt space."""
    new_spaces = []

    for k, s in hp_space.items():

        new_spaces.append(skopt_space_from_hp_space(s, k))

    return new_spaces


def skopt_space_from_hp_space(hp_space, prior_name=None):
    """Converts on hp space into a corresponding skopt space."""
    if is_choice(hp_space):
        skopt_space = to_categorical(hp_space, prior_name=prior_name)
    elif any([i in hp_space.__str__() for i in ['loguniform', 'quniform', 'qloguniform', 'uniform']]):
        skopt_space = to_real(hp_space, prior_name=prior_name)
    elif 'randint' in hp_space.__str__():
        skopt_space = to_int(hp_space, prior_name=prior_name)
    else:
        raise NotImplementedError

    return skopt_space


def to_categorical(_space, prior_name=None):
    """converts an hp space into a Dimension object."""
    cats = []
    inferred_name = None
    for arg in _space.pos_args:
        if hasattr(arg, '_obj') and arg.pure:
            cats.append(arg._obj)
        elif arg.name == 'hyperopt_param' and len(arg.pos_args)>0:
            for a in arg.pos_args:
                if a.name == 'literal' and a.pure:
                    inferred_name = a._obj

    prior_name = verify_name(prior_name, inferred_name)

    if len(cats) == 0:
        raise NotImplementedError
    return Categorical(categories=cats, name=prior_name)


def verify_name(prior_name, inferred_name):
    """Verfies that the given/prior name matches with the inferred name.
    """
    if prior_name is None:
        prior_name = inferred_name
    else:
        assert prior_name == inferred_name, f"""given name {prior_name} does not mach with 
                                                inferred name {inferred_name}"""
    return prior_name


def to_int(_space, prior_name=None):
    """converts an hp.randint into a Dimension object"""
    inferred_name = None
    limits = None
    for arg in _space.pos_args:
        if arg.name == 'literal' and len(arg.named_args) == 0:
            inferred_name = arg._obj
        elif len(arg.named_args) == 2 and arg.name == 'randint':
            limits = {}
            for a in arg.named_args:
                limits[a[0]] = a[1]._obj

        elif len(arg.pos_args) == 2 and arg.name == 'randint':
            # high and low are not named args
            _limits = []
            for a in arg.pos_args:
                _limits.append(a._obj)
            limits = {'high': np.max(_limits), 'low': np.min(_limits)}
        else:
            raise NotImplementedError

    prior_name = verify_name(prior_name, inferred_name)

    if limits is None:
        raise NotImplementedError
    return Integer(low=limits['low'], high=limits['high'], name=prior_name)


def to_real(_space, prior_name=None):
    """converts an an hp space to real. """
    inferred_name = None
    limits = None
    prior = None
    allowed_names = ['uniform', 'loguniform', 'quniform', 'loguniform', 'qloguniform']
    for arg in _space.pos_args:
        if len(arg.pos_args) > 0:
            for a in arg.pos_args:
                if a.name == 'literal' and len(arg.named_args) == 0:
                    inferred_name = a._obj
                elif a.name in allowed_names and len(a.named_args) == 2:
                    prior = a.name
                    limits = {}
                    for _a in a.named_args:
                        limits[_a[0]] = _a[1]._obj
                elif a.name in allowed_names and len(a.pos_args) == 2:
                    prior = a.name
                    _limits = []
                    for _a in a.pos_args:
                        _limits.append(_a._obj)
                    limits = {'high': np.max(_limits), 'low': np.min(_limits)}

    prior_name = verify_name(prior_name, inferred_name)

    # prior must be inferred because hp spaces always have prior
    if limits is None or prior is None or 'high' not in limits:
        raise NotImplementedError

    return Real(low=limits['low'], high=limits['high'], prior=prior, name=prior_name)


def loss_histogram(losses,  # array like
                   xlabel='objective_fn',
                   ylabel='Frequency',
                   save=True,
                   fname="histogram.png"):

    plt.hist(losses)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    if save:
        plt.savefig(fname, dpi=300, bbox_inches='tight')
    else:
        plt.show()


def plot_hyperparameters(
        trials,
        save=True,
        fontsize=10,
        colorize_best=None,
        columns=5,
        arrange_by_loss=False,
        fname='parameters_selection_plot.png'
):
    """Copying from hyperopt because original hyperopt does not allow saving the plot."""

    idxs, vals = miscs_to_idxs_vals(trials.miscs)
    losses = trials.losses()
    finite_losses = [y for y in losses if y not in (None, float("inf"))]
    asrt = np.argsort(finite_losses)
    if colorize_best is not None:
        colorize_thresh = finite_losses[asrt[colorize_best + 1]]
    else:
        # -- set to lower than best (disabled)
        colorize_thresh = finite_losses[asrt[0]] - 1

    loss_min = min(finite_losses)
    loss_max = max(finite_losses)
    print("finite loss range", loss_min, loss_max, colorize_thresh)

    loss_by_tid = dict(zip(trials.tids, losses))

    def color_fn_bw(lossval):
        if lossval in (None, float("inf")):
            return 1, 1, 1
        else:
            t = (lossval - loss_min) / (loss_max - loss_min + 0.0001)
            if lossval < colorize_thresh:
                return 0.0, 1.0 - t, 0.0  # -- red best black worst
            else:
                return t, t, t  # -- white=worst, black=best

    all_labels = list(idxs.keys())
    titles = all_labels
    order = np.argsort(titles)

    C = min(columns, len(all_labels))
    R = int(np.ceil(len(all_labels) / float(C)))

    for plotnum, varnum in enumerate(order):
        label = all_labels[varnum]
        plt.subplot(R, C, plotnum + 1)

        # hide x ticks
        ticks_num, _ = plt.xticks()
        plt.xticks(ticks_num, [""] * len(ticks_num))

        dist_name = label

        if arrange_by_loss:
            x = [loss_by_tid[ii] for ii in idxs[label]]
        else:
            x = idxs[label]
        if "log" in dist_name:
            y = np.log(vals[label])
        else:
            y = vals[label]
        plt.title(titles[varnum], fontsize=fontsize)
        c = list(map(color_fn_bw, [loss_by_tid[ii] for ii in idxs[label]]))
        if len(y):
            plt.scatter(x, y, c=c)
        if "log" in dist_name:
            nums, _ = plt.yticks()
            plt.yticks(nums, ["%.2e" % np.exp(t) for t in nums])

    if save:
        plt.savefig(fname, dpi=300, bbox_inches='tight')
    else:
        plt.show()


def post_process_skopt_results(skopt_results, results, opt_path):

    skopt_plots(skopt_results, pref=opt_path)

    if 'folder' in list(results.items())[0]:
        clear_weights(results=results, opt_dir=opt_path)
    else:
        clear_weights(results=results, opt_dir=opt_path)
        clear_weights(opt_dir=opt_path)

    return


def save_skopt_results(skopt_results, opt_path):

    fname = os.path.join(opt_path, 'gp_parameters')

    sr_res = SerializeSKOptResults(skopt_results)

    try:  # adding .hpo extension so that no other file/folder exists with same name
        fname = os.path.join(opt_path, os.path.basename(opt_path) + ".hpo")
        dump(skopt_results, fname)
    except PicklingError:
        print("could not pickle results")

    try:
        with open(fname + '.json', 'w') as fp:
            json.dump(sr_res.serialized_results, fp, sort_keys=True, indent=4)
    except TypeError:
        with open(fname + '.json', 'w') as fp:
            json.dump(str(sr_res.serialized_results), fp, sort_keys=True, indent=4)

    return


def skopt_plots(search_result, pref=os.getcwd(), threshold=20):

    if len(search_result.x) < threshold:  # it takes forever if parameters are > 20
        plt.close('all')
        _ = plot_evaluations(search_result)
        plt.savefig(os.path.join(pref, 'evaluations'), dpi=400, bbox_inches='tight')

    if len(search_result.x) < threshold:
        if search_result.space.n_dims == 1:
            pass
        else:
            plt.close('all')
            _ = plot_objective(search_result)
            plt.savefig(os.path.join(pref, 'objective'), dpi=400, bbox_inches='tight')

    convergence(search_result.func_vals)
    plt.savefig(os.path.join(pref, 'convergence'), dpi=300, bbox_inches='tight')
    convergence(search_result.func_vals, show_original=True)
    plt.savefig(os.path.join(pref, 'convergence_original'), dpi=300, bbox_inches='tight')
    return


def convergence(func_vals, color=None, show_original=False):

    _, ax = plt.subplots()
    ax.grid()

    n_calls = len(func_vals)
    mins = [np.min(func_vals[:i])
            for i in range(1, n_calls + 1)]

    if show_original:
        data = func_vals
    else:
        data = mins

    plot(data,
         color=color,
         marker=".", markersize=12, lw=2,
         title="Convergence plot",
         xlabel="Number of calls $n$",
         ylabel=r"$\min f(x)$ after $n$ calls",
         show=False,
         ax=ax)

    return

class SerializeSKOptResults(object):
    """
    This class has two functions
      - converts everything in skopt results into python native types so that these results can be saved in readable
        json files.
      - Store as much attributes in in serialized form that a skopt `search_result` object can be generated from it
        which then can be used to regenerate all hyper-parameter optimization related plots.

     skopt_results is a dictionary which contains following keys
     x: list, list of parameters being optimized
     fun: float, final value of objective function
     func_vals: numpy array, of length equal to number of iterations
     x_iters: list of lists, outer list is equal to number of iterations and each inner list is equal number of parameters
              being optimized
     models: list of models, where a model has following attributes
             - noise: str
             - kernel: skopt.learning.gaussian_process.kernels.Sum, it has following 2 attributes
                 - k1: skopt.learning.gaussian_process.kernels.Product, it has following attributes
                     -k1: skopt.learning.gaussian_process.kernels.ConstantKernel
                         - constant_value: flaot
                         - constant_value_bounds: tuple of floats
                     -k2: skopt.learning.gaussian_process.kernels.Matern
                         - length_scale: numpy ndarray
                         - length_scale_bounds: list of floats
                         - nu: float
                 - k2: skopt.learning.gaussian_process.kernels.WhiteKernel
                     - noise_level: float
                     - noise_level_bounds: tuple of floats
             - alpha: float
             - optimizer: str
             - n_restarts_optimizer: int
             - normalize_y: bool
             - copy_X_train: bool
             - random_state: int
             - kernel_: skopt.learning.gaussian_process.kernels.Sum
             - _rng: numpy.random.mtrand.RandomState
             - n_features_in_: int
             - _y_train_mean: np.float64
             - _y_train_std: np.float64
             - X_train_: numpy array
             - y_train_: numpy array
             - log_marginal_likelihood_value_: numpy array
             - L_: numpy array
             - _K_inv: NoneType
             - alpha_: numpy array
             - noise_: np.float64
             - K_inv_: numpy array
             - y_train_std_: np.float64
             - y_train_mean_: np.float64
     space: skopt.space.space.Space, parameter spaces
     random_state: numpy.random.mtrand.RandomState
     specs: dict,  specs of each iteration. It has following keys
         - args: dict, which has following keys
             - func: function
             - dimensions: skopt.space.space.Space
             - base_estimator: skopt.learning.gaussian_process.gpr.GaussianProcessRegressor, which has following attributes
                 - noise: str
                 - kernel: skopt.learning.gaussian_process.kernels.Product
                     - k1: skopt.learning.gaussian_process.kernels.ConstantKernel, which has following attributes
                         - constant_value: flaot
                         - constant_value_bounds: tuple of floats
                     - k2: skopt.learning.gaussian_process.kernels.Matern, which has following attributes
                         - length_scale: numpy ndarray
                         - length_scale_bounds: list of floats
                         - nu: float
                 - alpha: float
                 - optimizer: str
                 - n_restarts_optimizer: int
                 - normalize_y: bool
                 - copy_X_train: bool
                 - random_state: int
             - n_cals: int
             - n_random_starts: NoneType
             - n_initial_points: str
             - initial_point_generator: str
             - acq_func: str
             - acq_optimizer: str
             - x0: list
             - y0: NoneType
             - random_state: numpy.random.mtrand.RandomState
             - verbose: book
             - callback: NoneType
             - n_points: int
             - n_restarts_optimizer: int
             - xi: float
             - kappa: float
             - n_jobs: int
             - model_queue_size: NoneType
         - function: str
     """
    def __init__(self, results: dict):
        self.results = results
        self.iters = len(results['func_vals'])
        self.paras = len(results['x'])
        self.serialized_results = {}

        for key in results.keys():
            self.serialized_results[key] = getattr(self, key)()

    def x(self):

        return self.para_list(self.results['x'])

    def para_list(self, x):
        """Serializes list of parameters"""
        _x = []
        for para in x:
            _x.append(Jsonize(para)())
        return _x

    def x0(self):
        _x0 = []

        __xo = self.results['specs']['args']['x0']

        if __xo is not None:
            for para in __xo:
                if isinstance(para, list):
                    _x0.append(self.para_list(para))
                else:
                    _x0.append(Jsonize(para)())
        return _x0

    def y0(self):
        __y0 = self.results['specs']['args']['y0']
        if __y0 is None:
            return __y0
        if isinstance(__y0, list):
            _y0 = []
            for y in self.results['specs']['args']['y0']:
                _y0.append(Jsonize(y)())
            return _y0

        return Jsonize(self.results['specs']['args']['y0'])()

    def fun(self):
        return float(self.results['fun'])

    def func_vals(self):
        return [float(i) for i in self.results['func_vals']]

    def x_iters(self):
        out_x = []
        for i in range(self.iters):
            x = []
            for para in self.results['x_iters'][i]:
                x.append(Jsonize(para)())

            out_x.append(x)

        return out_x

    def space(self):
        raum = {}
        for sp in self.results['space'].dimensions:
            if sp.__class__.__name__ == 'Categorical':
                _raum = {k: Jsonize(v)() for k, v in sp.__dict__.items() if k in ['categories', 'transform_',
                                                                                  'prior', '_name']}
                _raum.update({'type': 'Categorical'})
                raum[sp.name] = _raum

            elif sp.__class__.__name__ == 'Integer':
                _raum = {k: Jsonize(v)() for k, v in sp.__dict__.items() if
                                       k in ['low', 'transform_', 'prior', '_name', 'high', 'base',
                                             'dtype', 'log_base']}
                _raum.update({'type': 'Integer'})
                raum[sp.name] = _raum

            elif sp.__class__.__name__ == 'Real':
                _raum = {k: Jsonize(v)() for k, v in sp.__dict__.items() if
                                       k in ['low', 'transform_', 'prior', '_name', 'high', 'base', 'dtype',
                                             'log_base']}
                _raum.update({'type': 'Real'})
                raum[sp.name] = _raum

        return raum

    def random_state(self):
        return str(self.results['random_state'])

    def kernel(self, k):
        """Serializes Kernel"""
        if k.__class__.__name__ == "Product":
            return self.prod_kernel(k)

        if k.__class__.__name__ == "Sum":
            return self.sum_kernel(k)
        # default scenario, just converts it to string
        return str(k)

    def prod_kernel(self, k):
        """Serializes product kernel"""
        kernel = {}
        for _k, v in k.__dict__.items():

            kernel[_k] = self.singleton_kernel(v)

        return {"ProductKernel": kernel}

    def sum_kernel(self, k):
        """Serializes sum kernel"""
        kernel = {}

        for _k, v in k.__dict__.items():

            if v.__class__.__name__ == "Product":
                kernel[_k] = self.prod_kernel(v)
            else:
                kernel[_k] = self.singleton_kernel(v)

        return {"SumKernel": kernel}

    def singleton_kernel(self, k):
        """Serializes Kernels such as  Matern, White, Constant Kernels"""
        return {k: Jsonize(v)() for k, v in k.__dict__.items()}

    def specs(self):
        _specs = {}

        _specs['function'] = self.results['specs']['function']

        args = {}

        args['func'] = str(self.results['specs']['args']['func'])

        args['dimensions'] = self.space()

        be = self.results['specs']['args']['base_estimator']
        b_e = {k: Jsonize(v)() for k, v in be.__dict__.items() if
                                       k in ['noise', 'alpha', 'optimizer', 'n_restarts_optimizer', 'normalize_y',
                                             'copy_X_train', 'random_state']}
        b_e['kernel'] = self.kernel(be.kernel)

        args['base_estimator'] = b_e

        for k,v in self.results['specs']['args'].items():
            if k in ['n_cals', 'n_random_starts', 'n_initial_points', 'initial_point_generator', 'acq_func',
                     'acq_optimizer', 'verbose', 'callback', 'n_points', 'n_restarts_optimizer', 'xi', 'kappa',
                     'n_jobs', 'model_queue_size']:
                args[k] = Jsonize(v)()

        args['x0'] = self.x0()
        args['y0'] = self.y0()

        _specs['args'] = args
        return _specs

    def models(self):

        mods = []
        for model in self.results['models']:
            mod = {k: Jsonize(v)() for k, v in model.__dict__.items() if k in [
                'noise','alpha', 'optimizer', 'n_restarts_optimizer', 'normalize_y', 'copy_X_train', 'random_state',
                '_rng', 'n_features_in', '_y_tain_mean', '_y_train_std', 'X_train', 'y_train',
                'log_marginal_likelihood', 'L_', 'K_inv', 'alpha', 'noise_', 'K_inv_', 'y_train_std_', 'y_train_mean_']}

            mod['kernel'] = self.kernel(model.kernel)
            mods.append({model.__class__.__name__: mod})

        return mods


def to_skopt_space(x):
    """converts the space x into skopt compatible space"""
    if isinstance(x, list):
        if all([isinstance(s, Dimension) for s in x]):
            _space = Space(x)
        elif len(x) == 1 and isinstance(x[0], tuple):
            if len(x[0]) == 2:
                if 'int' in x[0][0].__class__.__name__:
                    _space = Integer(low=x[0][0], high=x[0][1])
                elif 'float' in x[0][0].__class__.__name__:
                    _space = Integer(low=x[0][0], high=x[0][1])
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError
        elif all([s.__class__.__name__== "Apply" for s in x]):
            _space = Space([skopt_space_from_hp_space(v) for v in x])
        else:
            raise NotImplementedError
    elif isinstance(x, dict):  # todo, in random, should we build Only Categorical space?
        space_ = []
        for k, v in x.items():
            if isinstance(v, list):
                s = space_from_list(v, k)
            elif isinstance(v, Dimension):
                # it is possible that the user has not specified the name so assign the names
                # because we have keys.
                if v.name is None or v.name.startswith('real_') or v.name.startswith('integer_'):
                    v.name = k
                s = v
            elif v.__class__.__name__== "Apply" or 'rv_frozen' in v.__class__.__name__:
                s = skopt_space_from_hp_space(v, k)
            elif isinstance(v, tuple):
                s = Categorical(v, name=k)
            elif isinstance(v, np.ndarray):
                s = Categorical(v.tolist(), name=k)
            else:
                raise NotImplementedError(f"unknown type {v}, {type(v)}")
            space_.append(s)

        # todo, why converting to Space
        _space = Space(space_) if len(space_) > 0 else None
    elif 'rv_frozen' in x.__class__.__name__ or x.__class__.__name__== "Apply":
        _space = Space([skopt_space_from_hp_space(x)])
    else:
        raise NotImplementedError(f"unknown type {x}, {type(x)}")
    return _space


def scatterplot_matrix_colored(params_names: list,
                               params_values: list,
                               best_accs: list,
                               blur=False,
                               save=True,
                               fname='scatter_plot.png'
                               ):
    """Scatterplot colored according to the Z values of the points.
    https://github.com/guillaume-chevalier/Hyperopt-Keras-CNN-CIFAR-100/blob/Vooban/AnalyzeTestHyperoptResults.ipynb
    """

    nb_params = len(params_values)
    best_accs = np.array(best_accs)
    norm = mpl.colors.Normalize(vmin=best_accs.min(), vmax=best_accs.max())

    fig, ax = plt.subplots(nb_params, nb_params, figsize=(16, 16))  # , facecolor=bg_color, edgecolor=fg_color)

    for i in range(nb_params):
        p1 = params_values[i]
        for j in range(nb_params):
            p2 = params_values[j]

            axes = ax[i, j]
            # Subplot:
            if blur:
                axes.scatter(p2, p1, s=400, alpha=.1,
                             c=best_accs, cmap='viridis', norm=norm)
                axes.scatter(p2, p1, s=200, alpha=.2,
                             c=best_accs, cmap='viridis', norm=norm)
                axes.scatter(p2, p1, s=100, alpha=.3,
                             c=best_accs, cmap='viridis', norm=norm)
            s = axes.scatter(p2, p1, s=15,
                             c=best_accs, cmap='viridis', norm=norm)

            # Labels only on side subplots, for x and y:
            if j == 0:
                axes.set_ylabel(params_names[i], rotation=0)
            else:
                axes.set_yticks([])

            if i == nb_params - 1:
                axes.set_xlabel(params_names[j], rotation=90)
            else:
                axes.set_xticks([])

    fig.subplots_adjust(right=0.82, top=0.95)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(s, cax=cbar_ax)

    plt.suptitle(
        'Mmatrix of tried values in the search space over different params, colored in function of best test accuracy')
    plt.show()


def take(n, iterable):
    """Return first n items of the iterable as a list
    https://stackoverflow.com/questions/7971618/return-first-n-keyvalue-pairs-from-dict
    """
    return list(islice(iterable, n))


def plot_convergences(opt_dir, what='val_loss', show_whole=True, show_min=False, **kwargs):

    plot_dir = os.path.join(opt_dir, "plots")
    if not os.path.exists(plot_dir):
        os.makedirs(plot_dir)

    max_len = kwargs.get('max_len', 500)
    show_top = kwargs.get('show_top', 3)
    ylim_bottom = kwargs.get('ylim_bottom', None)
    ylim_top = kwargs.get('ylim_top', None)
    font_size = kwargs.get('font_size', 16)
    ylabel = kwargs.get('ylabel', 'Validation MSE')
    leg_pos = kwargs.get('leg_pos', 'upper right')
    # style = kwargs.get('style', 'ggplot')

    models = []
    for f in os.listdir(opt_dir):
        if os.path.isdir(os.path.join(opt_dir, f)):
            models.append(f)

    val_loss = pd.DataFrame()
    default = pd.Series(np.full(max_len, 0.0))
    min_val_loss = {}

    for mod in models:
        loss_fname = os.path.join(os.path.join(opt_dir, str(mod) + "\\losses.csv"))
        if os.path.exists(loss_fname):
            losses = pd.read_csv(loss_fname)
            vl = losses[what]

            vl1 = default.add(vl, fill_value=None)
            vl1.name = mod

            min_val_loss[mod] = vl1.min()

            val_loss = pd.concat([val_loss, vl1], axis=1)

    # sort min_val_loss by value
    min_vl_sorted = {k: v for k, v in sorted(min_val_loss.items(), key=lambda item: item[1])}
    top_3 = take(show_top, min_vl_sorted)

    colors = {
        0: 'b',
        1: 'g',
        2: 'orange'
    }

    default = np.full(max_len, np.nan)

    plt.close('all')
    _, axis = plt.subplots()

    for k in min_vl_sorted.keys():

        val = val_loss[k]

        default[np.argmin(val.values)] = val.min()
        if k not in top_3:
            if show_whole:
                axis.plot(val, color='silver', linewidth=0.5,  label='_nolegend_')

            if show_min:
                axis.plot(default, '.', markersize=2.5, color='silver', label='_nolegend_')
        default = np.full(max_len, np.nan)

    for idx, k in enumerate(top_3):
        val = val_loss[k]

        if show_whole:
            axis.plot(val, color=colors[idx], linewidth=1,   label=f'Rank {idx+1}')

        default[np.argmin(val.values)] = val.min()
        if show_min:
            axis.plot(default, 'x', markersize=5.5, color=colors[idx], label='Rank ' + str(idx))

        default = np.full(max_len, np.nan)

    axis.legend(loc=leg_pos, fontsize=font_size)
    axis.set_yscale('log')
    axis.set_ylabel(ylabel, fontsize=font_size)
    axis.set_xlabel('Epochs', fontsize=font_size)

    if ylim_bottom is not None:
        axis.set_ylim(bottom=ylim_bottom)
    if ylim_top is not None:
        axis.set_ylim(top=ylim_top)

    _name = what
    _name += '_whole_loss_' if show_whole else ''
    _name += '_loss_points' if show_min else ''
    fname = os.path.join(plot_dir, _name)

    plt.savefig(fname, dpi=300, bbox_inches='tight')

    return


def to_skopt_as_dict(algorithm:str, backend:str, original_space)->dict:

    if backend == 'hyperopt':
        if original_space.__class__.__name__ == "Apply":
            _space = skopt_space_from_hp_space(original_space)
            _space = {_space.name: _space}
        elif isinstance(original_space, dict):
            _space = OrderedDict()
            for k, v in original_space.items():
                if v.__class__.__name__ == "Apply" or 'rv_frozen' in v.__class__.__name__:
                    _space[k] = skopt_space_from_hp_space(v)
                elif isinstance(v, Dimension):
                    _space[v.name] = v
                else:
                    raise NotImplementedError
        elif isinstance(original_space, list):
            if all([isinstance(s, Dimension) for s in original_space]):
                _space = OrderedDict({s.name: s for s in original_space})
            elif all([s.__class__.__name__== "Apply" for s in original_space]):
                d = [skopt_space_from_hp_space(v) for v in original_space]
                _space = OrderedDict({s.name: s for s in d})
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    elif backend == 'optuna':
        if isinstance(original_space, list):
            if all([isinstance(s, Dimension) for s in original_space]):
                _space = OrderedDict({s.name: s for s in original_space})
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    elif backend == 'skopt':
        sk_space = to_skopt_space(original_space)

        if isinstance(sk_space, Dimension):
            _space = {sk_space.name: sk_space}

        elif all([isinstance(s, Dimension) for s in sk_space]):
            _space = OrderedDict()
            for s in sk_space:
                _space[s.name] = s
        else:
            raise NotImplementedError

    elif backend == 'sklearn':
        if isinstance(original_space, list):
            if all([isinstance(s, Dimension) for s in original_space]):
                _space = OrderedDict({s.name: s for s in original_space})
            else:
                raise NotImplementedError
        elif isinstance(original_space, dict):
            _space = OrderedDict()
            for k, v in original_space.items():
                if isinstance(v, list):
                    s = space_from_list(v, k)
                elif isinstance(v, Dimension):
                    s = v
                elif isinstance(v, tuple) or isinstance(v, list):
                    s = Categorical(v, name=k)
                elif  v.__class__.__name__ == "Apply" or 'rv_frozen' in v.__class__.__name__:
                    if algorithm == 'random':
                        s = Real(v.kwds['loc'], v.kwds['loc'] + v.kwds['scale'], name=k, prior=v.dist.name)
                    else:
                        s = skopt_space_from_hp_space(v)
                else:
                    raise NotImplementedError(f"unknown type {v}, {type(v)}")
                _space[k] = s
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    return _space


def space_from_list(v: list, k: str):
    """Returns space of tyep "Dimension"
    """
    if len(v) > 2:
        if isinstance(v[0], int):
            s = Integer(grid=v, name=k)
        elif isinstance(v[0], float):
            s = Real(grid=v, name=k)
        else:
            s = Categorical(v, name=k)
    else:
        if isinstance(v[0], int):
            s = Integer(low=np.min(v), high=np.max(v), name=k)
        elif isinstance(v[0], float):
            s = Real(low=np.min(v), high=np.max(v), name=k)
        elif isinstance(v[0], str):
            s = Categorical(v, name=k)
        else:
            raise NotImplementedError
    return s


def plot_edf(y, num_points=100):
    """
    Plots the objective value empirical distribution function on hyperparameter
    optimization. Implementation is taken from optuna.
    """
    x = np.linspace(np.min(y), np.max(y), num_points)

    y_values = np.sum(y[:, np.newaxis] <= x, axis=0) / y.size

    _, ax = plt.subplots()
    ax.grid()
    ax = plot(x,
              y_values,
              '-',
              show=False,
              title="Empirical Distribution Function Plot",
              ylabel="Cumulative Probability",
              xlabel="Objective Value",
              ax=ax)

    return ax




from ai4water.backend import get_attributes

try:
    import torch
except ModuleNotFoundError:
    torch = None

LAYERS = {}

LOSSES = {}

OPTIMIZERS = {}

if torch is not None:
    LAYERS.update(get_attributes(torch, 'nn', case_sensitive=True))

    LOSSES.update({"MSE": torch.nn.MSELoss,
                   "mse": torch.nn.MSELoss,
                   "CROSSENTROPYLOSS": torch.nn.CrossEntropyLoss,
                   "L1Loss": torch.nn.L1Loss,
                   "NLLLoss": torch.nn.NLLLoss,
                   "HingeEmbeddingLoss": torch.nn.HingeEmbeddingLoss,
                   "MarginRankingLoss": torch.nn.MarginRankingLoss,
                   "TripletMarginLoss": torch.nn.TripletMarginLoss,
                   "KLDivLoss": torch.nn.KLDivLoss,

                   })
    OPTIMIZERS.update(get_attributes(torch, 'optim', case_sensitive=True))
